#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""

# Standard library imports
from typing import Union, Tuple, Optional
import math
import numpy as np

# Third party imports
import scine_database as db
import scine_utilities as utils

from warnings import warn
warn('This is deprecated, please import the queries from scine_database', DeprecationWarning, stacklevel=2)


def check_barrier_height(
        reaction: db.Reaction,
        database_manager: db.Manager,
        model: db.Model,
        structures: db.Collection,
        properties: db.Collection,
        max_barrier: float,
        min_barrier: float = -math.inf,
        barrierless_always_pass: bool = True)\
        -> Union[db.ElementaryStep, None]:
    """
    Check if the barrier of any elementary step associated to the given reaction is below max_barrier.


    Parameters
    ----------
    reaction : scine_database.Reaction
        The reaction.
    database_manager : scine_database.Manager
        The database manager. Required for accessing the elementary steps.
    model : scine_database.Model
        The electronic structure model.
    structures : scine_utils.Collection
        The structure collection.
    properties : scine_utils.Collection
        The property collection.
    max_barrier : float
        The max. barrier threshold.
    min_barrier : float
        The min. barrier threshold (default -inf)
    barrierless_always_pass : bool
        Always allow barrier-less reactions.

    Returns
    -------
    Union[scine_database.ElementaryStep, None]
        The first elementary step that fulfills this condition barrier < max_barrier is returned. If the condition
        is never met, None is returned.
    """
    for elementary_step in reaction.get_elementary_steps(database_manager):
        if not elementary_step.has_transition_state() and barrierless_always_pass:
            return elementary_step
        barrier = get_single_barrier_for_elementary_step(
            elementary_step, model, structures, properties
        )
        if barrier is None:
            continue
        if max_barrier > barrier > min_barrier:
            return elementary_step
    return None


def get_single_barrier_for_elementary_step(
    step: db.ElementaryStep,
    model: db.Model,
    structures: db.Collection,
    properties: db.Collection,
) -> Union[float, None]:
    """
    Gives the barrier height of a single elementary step (left to right) in kJ/mol. If available, the gibbs free energy
    ('gibbs_free_energy') barrier is returned. Otherwise the electronic energy ('electronic_energy') barrier is
    returned. Returns None if both energies are unavailable.

    Parameters
    ----------
    step : scine_database.ElementaryStep (Scine::Database::ElementaryStep)
        The elementary step we want the barrier height from
    model : scine_database.Model
        The model used to calculate the energies.
    structures : scine_database.Collection
        The structure collection.
    properties : scine_database.Collection
        The property collection.

    Returns
    -------
    Union[float, None]
        barrier in kJ/mol
    """
    gibbs = get_single_barrier_for_elementary_step_by_type(
        step, "gibbs_free_energy", model, structures, properties
    )
    if gibbs is not None:
        return gibbs
    else:
        return get_single_barrier_for_elementary_step_by_type(
            step, "electronic_energy", model, structures, properties
        )


def get_energy_change(
        step: db.ElementaryStep,
        energy_type: str,
        model: db.Model,
        structures: db.Collection,
        properties: db.Collection,
) -> Union[None, float]:
    """
    Get the energy difference between educts and products for an elementary step.

    Parameters
    ----------
    step : db.ElementaryStep
        The elementary step.
    energy_type : str
        The energy label.
    model : db.Model
        The electronic structure model.
    structures : db.Collection
        The structure collection.
    properties : db.Collection
        The property collection.

    Returns
    -------
    Union[float, None]
        The difference between products and reactant energy. Returns None, if the energy is unavailable.
    """
    reactant_energies = [
        get_energy_for_structure(db.Structure(reactant), energy_type, model, structures, properties)
        for reactant in step.get_reactants(db.Side.LHS)[0]
    ]
    reactant_energy: Union[None, float] = None if None in reactant_energies else sum(reactant_energies)  # type: ignore
    product_energies = [
        get_energy_for_structure(db.Structure(product), energy_type, model, structures, properties)
        for product in step.get_reactants(db.Side.RHS)[1]
    ]
    product_energy: Union[None, float] = None if None in product_energies else sum(product_energies)  # type: ignore
    if not product_energy or not reactant_energy:
        return None
    return product_energy - reactant_energy


def get_single_barrier_for_elementary_step_by_type(
    step: db.ElementaryStep,
    energy_type: str,
    model: db.Model,
    structures: db.Collection,
    properties: db.Collection,
) -> Union[float, None]:
    """
    Gives the barrier height of a single elementary step (left to right) in kJ/mol for the specified energy type.
    Returns None if the energy type is not available.

    Parameters
    ----------
    step : scine_database.ElementaryStep (Scine::Database::ElementaryStep)
        The elementary step we want the barrier height from
    energy_type : str
        The name of the energy property such as 'electronic_energy' or 'gibbs_free_energy'
    model : scine_database.Model
        The model used to calculate the energies.
    structures : scine_database.Collection
        The structure collection.
    properties : scine_database.Collection
        The property collection.

    Returns
    -------
    Union[float, None]
        barrier in kJ/mol
    """
    reactant_energy = 0.0
    for reactant in step.get_reactants(db.Side.LHS)[0]:
        ret = get_energy_for_structure(db.Structure(reactant), energy_type, model, structures, properties)
        if ret is None:
            return None
        else:
            reactant_energy += ret
    if not step.has_transition_state():
        product_energy = 0.0
        for product in step.get_reactants(db.Side.RHS)[1]:
            ret = get_energy_for_structure(db.Structure(product), energy_type, model, structures, properties)
            if ret is None:
                return None
            else:
                product_energy += ret
        if product_energy > reactant_energy:
            return (product_energy - reactant_energy) * utils.KJPERMOL_PER_HARTREE
        else:
            return 0.0
    ts = db.Structure(step.get_transition_state())
    ts_energy = get_energy_for_structure(ts, energy_type, model, structures, properties)
    if ts_energy is None:
        return None
    else:
        return (ts_energy - reactant_energy) * utils.KJPERMOL_PER_HARTREE


def get_energy_for_structure(
    structure: db.Structure,
    prop_name: str,
    model: db.Model,
    structures: db.Collection,
    properties: db.Collection,
) -> Union[float, None]:
    """
    Gives energy value depending on demanded property. If the property does not exit, None is returned.

    Parameters
    ----------
    structure : scine_database.Structure (Scine::Database::Structure)
        The structure we want the energy from
    prop_name : str
        The name of the energy property such as 'electronic_energy' or 'gibbs_free_energy'
    model : scine_database.Model
        The model used to calculate the energies.
    structures : scine_database.Collection
        The structure collection.
    properties : scine_database.Collection
        The property collection.

    Returns
    -------
    Union[float, None]
        energy value in Hartree
    """
    structure.link(structures)
    structure_properties = structure.query_properties(prop_name, model, properties)
    if not structure_properties:
        return None
    # pick last property if multiple
    prop = db.NumberProperty(structure_properties[-1])
    prop.link(properties)
    return prop.get_data()


def get_energy_sum_of_elementary_step_side(step: db.ElementaryStep, side: db.Side, energy_type: str, model: db.Model,
                                           structures: db.Collection, properties: db.Collection) -> Union[float, None]:
    """
    Gives the total energy in atomic units of the given side of the step. Returns None if the energy type is
    not available.

    Parameters
    ----------
    step : scine_database.ElementaryStep (Scine::Database::ElementaryStep)
        The elementary step we want the energy from
    side : scine_database.Side (Scine::Database::Side)
        The side we want the side from
    energy_type : str
        The name of the energy property such as 'electronic_energy' or 'gibbs_free_energy'
    model : scine_database.Model
        The model used to calculate the energies.
    structures : scine_database.Collection
        The structure collection.
    properties : scine_database.Collection
        The property collection.

    Returns
    -------
    Union[float, None]
        Energy in hartree
    """
    if side == db.Side.BOTH:
        raise RuntimeError("The energy sum of both sides of a step is not supported.")
    index = 0 if side == db.Side.LHS else 1
    energies = [
        get_energy_for_structure(db.Structure(reactant), energy_type, model, structures, properties)
        for reactant in step.get_reactants(side)[index]
    ]
    return None if None in energies else sum(energies)  # type: ignore


def get_elementary_step_with_min_ts_energy(reaction: db.Reaction,
                                           energy_type: str,
                                           model: db.Model,
                                           elementary_steps: db.Collection,
                                           structures: db.Collection,
                                           properties: db.Collection,
                                           max_barrier: float,
                                           min_barrier=-math.inf,
                                           structure_model: Optional[db.Model] = None) -> Optional[db.ElementaryStep]:
    """
    Gets the elementary step ID with the lowest energy of the corresponding transition state of a reaction.


    Parameters
    ----------
    reaction : db.Reaction
        The reaction for which the elementary steps shall be analyzed.
    energy_type : str
        The name of the energy property such as 'electronic_energy' or 'gibbs_free_energy'
    model : scine_database.Model
        The model used to calculate the energies.
    elementary_steps : db.Collection
        The elementary step collection.
    structures : scine_database.Collection
        The structure collection.
    properties : scine_database.Collection
        The property collection.
    max_barrier : float
        The maximum allowed barrier
    min_barrier : float
        The minimum allowed barrier, default is negative infinity
    structure_model : Optional[db.Model]
        The model of the transition state. If None, the model of the transition state is not checked.

    Returns
    -------
    Union[db.ElementaryStep, None]
        The elementary step with the lowest TS energy.
    """
    lowest_ts_energy = np.inf
    es_id_with_lowest_ts = None
    # # # Loop over elementary steps
    for es_id in reaction.get_elementary_steps():
        es = db.ElementaryStep(es_id, elementary_steps)
        # # # Type check elementary step and break if barrierless
        if es.get_type() == db.ElementaryStepType.BARRIERLESS:
            first_structure_lhs = db.Structure(es.get_reactants()[0][0], structures)
            if structure_model is not None and first_structure_lhs.get_model() != structure_model:
                continue
            # # # Energy Check for minima
            first_structure_lhs_energy = get_energy_for_structure(first_structure_lhs, energy_type, model,
                                                                  structures, properties)
            if first_structure_lhs_energy is None:
                continue
            es_id_with_lowest_ts = es_id
            break
        ts = db.Structure(es.get_transition_state(), structures)
        if structure_model is not None and ts.get_model() != structure_model:
            continue

        # # # Costly safety check that barrier as well as ts_energy exist for this model and energy type
        ts_energy = get_energy_for_structure(
            ts, energy_type, model, structures, properties)
        barriers = get_barriers_for_elementary_step_by_type(es, energy_type, model, structures, properties)
        if None in barriers or ts_energy is None:
            continue
        # Check min/max barrier are correct, check next elementary step
        if any(barrier > max_barrier or barrier < min_barrier for barrier in barriers):  # type: ignore
            continue
        # # # Comparison with current lowest energy
        if ts_energy < lowest_ts_energy:
            es_id_with_lowest_ts = es_id
            lowest_ts_energy = ts_energy
    if es_id_with_lowest_ts is not None:
        return db.ElementaryStep(es_id_with_lowest_ts, elementary_steps)
    else:
        return None


def get_barriers_for_elementary_step_by_type(step: db.ElementaryStep, energy_type: str, model: db.Model,
                                             structures: db.Collection, properties: db.Collection)\
        -> Union[Tuple[float, float], Tuple[None, None]]:
    """
    Gives the forward and backward barrier height of a single elementary step (left to right) in kJ/mol for the
    specified energy type. Returns None if the energy type is not available.

    Parameters
    ----------
    step : scine_database.ElementaryStep (Scine::Database::ElementaryStep)
        The elementary step we want the barrier height from
    energy_type : str
        The name of the energy property such as 'electronic_energy' or 'gibbs_free_energy'
    model : scine_database.Model
        The model used to calculate the energies.
    structures : scine_database.Collection
        The structure collection.
    properties : scine_database.Collection
        The property collection.

    Returns
    -------
    Union[Tuple[float, float], Tuple[None, None]]
        Barriers in kJ/mol
    """
    reactant_energy = get_energy_sum_of_elementary_step_side(step, db.Side.LHS, energy_type, model,
                                                             structures, properties)
    if reactant_energy is None:
        return None, None
    product_energy = get_energy_sum_of_elementary_step_side(step, db.Side.RHS, energy_type, model,
                                                            structures, properties)
    if product_energy is None:
        return None, None
    if step.get_type() == db.ElementaryStepType.BARRIERLESS:
        if product_energy > reactant_energy:
            return (product_energy - reactant_energy) * utils.KJPERMOL_PER_HARTREE, 0.0
        else:
            return 0.0, (reactant_energy - product_energy) * utils.KJPERMOL_PER_HARTREE
    ts = db.Structure(step.get_transition_state())
    ts_energy = get_energy_for_structure(ts, energy_type, model, structures, properties)
    if ts_energy is None:
        return None, None
    lhs_barrier = (ts_energy - reactant_energy) * utils.KJPERMOL_PER_HARTREE
    rhs_barrier = (ts_energy - product_energy) * utils.KJPERMOL_PER_HARTREE
    return lhs_barrier, rhs_barrier



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""
"""
Utility functions for conversions between molecule/atom representations.
"""

from vtk import vtkPeriodicTable, vtkAtom, vtkMolecule
from typing import List, Tuple, cast, Optional
import numpy as np
import scine_utilities as su


def atom_to_tuple(atom: vtkAtom) -> Tuple[str, Tuple[float, float, float]]:
    """
    Returns an atomic symbol and a tuple of atom positions.
    """
    symbol = vtkPeriodicTable().GetSymbol(atom.GetAtomicNumber())
    position = atom.GetPosition()

    return cast(
        Tuple[str, Tuple[float, float, float]],
        (symbol, (position.GetX(), position.GetY(), position.GetZ())),
    )


def molecule_to_list_of_atoms(
    molecule: vtkMolecule,
) -> List[Tuple[str, Tuple[float, float, float]]]:
    """
    Convert vtkMolecule to list of atoms.
    """
    atoms = list()

    for atom_index in range(molecule.GetNumberOfAtoms()):
        atoms.append(atom_to_tuple(molecule.GetAtom(atom_index)))

    return atoms


def molecule_to_atom_collection(molecule: vtkMolecule) -> su.AtomCollection:
    """
    Convert list of atoms to su.AtomCollection.
    The method gets the molecule in Angstrom units and returns the AtomCollection in Bohr units.
    """
    atom_list = molecule_to_list_of_atoms(molecule)
    if not atom_list:
        return su.AtomCollection()

    elements = list()
    positions = list()

    for symbol, position in atom_list:
        elements.append(su.ElementInfo.element_from_symbol(symbol))
        positions.append(su.BOHR_PER_ANGSTROM * np.array(position))

    return su.AtomCollection(elements, np.array(positions))


def molecule_to_bond_order_collection(molecule: vtkMolecule) -> su.BondOrderCollection:
    bo_collection = su.BondOrderCollection(molecule.GetNumberOfAtoms())

    n_bonds = molecule.GetNumberOfBonds()
    for n in range(n_bonds):
        bond = molecule.GetBond(n)
        i = bond.GetBeginAtomId()
        j = bond.GetEndAtomId()
        order = bond.GetOrder()
        bo_collection.set_order(i, j, order)

    return bo_collection


def atom_collection_to_molecule(atom_collection: su.AtomCollection) -> vtkMolecule:
    molecule = vtkMolecule()
    for atom in atom_collection:
        p = atom.position * su.ANGSTROM_PER_BOHR
        z = su.ElementInfo.Z(atom.element)
        molecule.AppendAtom(z, *p)
    return molecule


def convert_gradients(
    gradients: np.ndarray,
    boost_factor: float = 0.1400142601462408,
    trust_radius: float = 0.2
) -> None:
    """
    Convert gradients from hartree/bohr to hartree/angstrom.
    Dampen, if max displacement is > trust_radius (default 1 angstrom)
    to avoid shooting around nuclei.
    """
    gradients *= boost_factor * su.BOHR_PER_ANGSTROM
    max_coefficient = np.max(np.abs(gradients))
    if max_coefficient > trust_radius:
        gradients *= trust_radius / max_coefficient


def apply_gradients(
    molecule: vtkMolecule,
    gradients: np.ndarray,
    bonds: Optional[np.ndarray] = None,
    mouse_picked_atom_ids: Optional[List[int]] = None,
    haptic_picked_atom_id: Optional[int] = None,
) -> None:
    """
    Apply gradients to molecule.
    """
    n_atoms = molecule.GetNumberOfAtoms()
    for atom_index in range(n_atoms):
        if (
            mouse_picked_atom_ids is not None
            and (
                haptic_picked_atom_id is None
                or haptic_picked_atom_id in mouse_picked_atom_ids
            )
            and atom_index in mouse_picked_atom_ids
        ):
            continue
        if haptic_picked_atom_id is not None and atom_index == haptic_picked_atom_id:
            continue
        gradient = gradients[atom_index]
        atom = molecule.GetAtom(atom_index)
        position = atom.GetPosition()

        atom.SetPosition(
            position[0] - gradient[0],
            position[1] - gradient[1],
            position[2] - gradient[2],
        )
    if not isinstance(bonds, np.ndarray):
        bond_collection = su.BondDetector.detect_bonds(molecule_to_atom_collection(molecule))
        bonds = bond_collection.matrix.toarray()

    for i in range(n_atoms - 1):
        for j in range(i + 1, n_atoms):
            bond_id = molecule.GetBondId(i, j)
            if bond_id == -1:
                molecule.AppendBond(i, j, int(round(bonds[i, j])))
            else:
                molecule.SetBondOrder(bond_id, int(round(bonds[i, j])))


def times_bohr_per_angstrom(atom_positions: List[float]) -> List[float]:
    """
    Convert atom positions from bohr to angstrom.
    """
    return cast(
        List[float],
        [
            atom_positions[0] * su.BOHR_PER_ANGSTROM,
            atom_positions[1] * su.BOHR_PER_ANGSTROM,
            atom_positions[2] * su.BOHR_PER_ANGSTROM,
        ],
    )


def times_angstrom_per_bohr(atom_positions: List[float]) -> List[float]:
    """
    Convert atom positions from angstrom to bohr.
    """
    return cast(
        List[float],
        [
            atom_positions[0] * su.ANGSTROM_PER_BOHR,
            atom_positions[1] * su.ANGSTROM_PER_BOHR,
            atom_positions[2] * su.ANGSTROM_PER_BOHR,
        ],
    )


def maximum_vdw_radius(molecule: vtkMolecule) -> float:
    """
    Returns the maximum VDW radius of all the atoms in the molecule.
    """
    table = vtkPeriodicTable()
    return max(
        (
            cast(float, table.GetVDWRadius(molecule.GetAtom(id).GetAtomicNumber()))
            for id in range(molecule.GetNumberOfAtoms())
        ),
        default=0.0,
    )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""
import numpy as np
from typing import Any, Optional, Tuple, TYPE_CHECKING
from matplotlib.figure import Figure
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg

import scine_utilities as utils

from scine_heron.utilities import (
    color_axis,
    color_figure,
    get_font,
    get_primary_line_color,
    get_primary_light_color,
)
from scine_heron.toolbar.io_toolbar import HeronToolBar
from scine_heron.database.energy_diagram import EnergyDiagram

from PySide2.QtWidgets import (
    QWidget,
    QFileDialog,
    QHBoxLayout,
)

if TYPE_CHECKING:
    Signal = Any
else:
    from PySide2.QtCore import Signal


class ReactionProfileWidget(QWidget):  # type: ignore
    def __init__(
        self,
        parent: Optional[QWidget] = None,  # pylint: disable=unused-argument
        width: float = 5,
        height: float = 3
    ):
        super(ReactionProfileWidget, self).__init__(parent)
        self.fig = Figure(figsize=(width, height))
        self.ax1 = self.fig.add_subplot(1, 1, 1)
        self.canvas = FigureCanvasQTAgg(self.fig)
        self.canvas.setMinimumHeight(150)
        self.canvas.setMinimumWidth(250)
        self._spline_order = 3
        self._spline: Optional[utils.bsplines.TrajectorySpline] = None
        self._prev_point = None

        self._invert = False

        self.plot_tool_bar = HeronToolBar(parent=self)
        self.plot_tool_bar.shortened_add_action('save_plot.png', "Save plot", "",
                                                self.save_svg)

        layout = QHBoxLayout()
        layout.addWidget(self.canvas)
        layout.addWidget(self.plot_tool_bar)
        self.setLayout(layout)
        self.clear_canvas()

    def update_canvas(self,
                      spline: Optional[utils.bsplines.TrajectorySpline] = None,
                      barriers: Optional[Tuple[float, float]] = None,
                      draw_signal: Optional[Signal] = None,
                      invert: bool = False,
                      has_transition_state: bool = False):
        if spline is None and\
           (barriers is None or any(barrier is None for barrier in barriers)):
            self.clear_canvas()
            self._prepare_canvas()
            self.ax1.set_title("No data to plot, check specified model.", get_font())
            return
        self._invert = invert
        self._prepare_canvas()
        # All energies are always relative to shown lhs
        lhs_energy = 0.0
        if barriers is not None and self._invert:
            barriers = barriers[::-1]

        if spline is None and barriers is not None:
            rhs_energy = barriers[0] - barriers[1]
            if has_transition_state:
                self._plot_plateaus(lhs_energy, rhs_energy, (barriers[0], 0.5))
            else:
                self._plot_plateaus(lhs_energy, rhs_energy)
        else:
            assert spline
            ts_position = spline.ts_position if not self._invert else 1.0 - spline.ts_position
            if barriers is not None:
                rhs_energy = barriers[0] - barriers[1]
                self._plot_plateaus(lhs_energy, rhs_energy, (barriers[0], ts_position))
            spline_lhs_energy, spline_ts_energy, spline_rhs_energy = self._plot_spline(spline)

        # Legend Plotting
        if barriers is None or any(barrier is None for barrier in barriers):
            self._plot_xlabel((spline_ts_energy - spline_lhs_energy) * utils.KJPERMOL_PER_HARTREE,
                              (spline_ts_energy - spline_rhs_energy) * utils.KJPERMOL_PER_HARTREE,
                              from_spline=True)
        else:
            self._plot_xlabel(barriers[0], barriers[1])  # (ts - start) * utils.KJPERMOL_PER_HARTREE

        # Final formatting
        self.fig.tight_layout()
        if draw_signal is not None and spline is not None:
            draw_signal.connect(self.draw_point)  # pylint: disable=no-member
            self.draw_point(0, 1)

        self.canvas.draw()

    def clear_canvas(self):
        self.fig.set_visible(False)
        self.plot_tool_bar.hide()
        self.canvas.draw()

    def _prepare_canvas(self) -> None:
        self.fig.set_visible(True)
        self.plot_tool_bar.show()
        self.ax1.cla()
        color_figure(self.fig)
        font = get_font()
        color_axis(self.ax1)
        self.ax1.set_title("Interpolated Reaction Path", font)
        self.ax1.set_ylabel("Electronic Energy in kJ/mol", font)
        self.ax1.set_xticks([], minor=[])

    def _plot_plateaus(self, lhs_energy: float, rhs_energy: float, lhs_barrier: Optional[Tuple[float, float]] = None):
        # Style Dictionaries
        diagram = EnergyDiagram("auto")
        space_for_level = 0.4  # assuming 3 plateaus
        diagram.dimension = space_for_level * 0.25
        diagram.space = space_for_level * 0.75
        # Color scheme
        # TODO adapt to dark mode
        # TODO: Maybe as general setting
        default_line_color = '#85929e'
        ts_level_color = '#995151'
        alpha = 0.75
        # Add Levels
        diagram.add_level(lhs_energy, position=0.0,
                          top_text="", bottom_text="", color=default_line_color, linewidth=4.5, alpha=alpha)
        if lhs_barrier is not None:
            diagram.add_level(lhs_barrier[0], position=(lhs_barrier[1] - 0.05) / space_for_level,
                              top_text="", bottom_text="", color=ts_level_color, linewidth=4.5, alpha=alpha)
            diagram.add_level(rhs_energy, position=0.9 / space_for_level, top_text="", bottom_text="",
                              color=default_line_color, linewidth=4.5, alpha=alpha)
            diagram.add_link(1, 2, linewidth=2.0, color=default_line_color, alpha=alpha)
        else:
            diagram.add_level(rhs_energy, top_text="", bottom_text="",
                              color=default_line_color, linewidth=4.5)
        # Add Link
        diagram.add_link(0, 1, linewidth=2.0, color=default_line_color)

        diagram.plot(ax=self.ax1)

    def _plot_spline(self, spline: utils.bsplines.TrajectorySpline) -> Tuple[float, float, float]:
        self._spline = spline
        spline_points = 1000
        spline_start_energy, _ = spline.evaluate(0.0, self._spline_order)
        spline_ts_energy, _ = spline.evaluate(spline.ts_position, self._spline_order)
        spline_end_energy, _ = spline.evaluate(1.0, self._spline_order)
        spline_x_points = np.array([i / spline_points for i in range(spline_points + 1)])
        spline_energies = np.array([(spline.evaluate(i / spline_points, self._spline_order)[0])
                                    * utils.KJPERMOL_PER_HARTREE for i in range(spline_points + 1)])
        # invert energy plot if required
        if self._invert:
            spline_x_points = 1 - spline_x_points[::-1]
            spline_energies = spline_energies[::-1]
            # Invert spline end point energies
            tmp_spline_start = spline_start_energy
            spline_start_energy = spline_end_energy
            spline_end_energy = tmp_spline_start
        # # # Rebase spline energies to start energy
        spline_energies -= spline_start_energy * utils.KJPERMOL_PER_HARTREE
        # Plot curve
        self.ax1.plot(spline_x_points, spline_energies,
                      color=get_primary_line_color(), zorder=2.0)
        return spline_start_energy, spline_ts_energy, spline_end_energy

    def _plot_xlabel(self, barrier_forward: float, barrier_backward: float, from_spline=False):
        label = "Reaction Coordinate\n" + "$\\Delta_{\\mathrm{R}} E_{el}$: " +\
                f" {(barrier_forward-barrier_backward):.1f}, " +\
                " $\\Delta^{\\ddag} E_{el}^{f}$: " + f" {barrier_forward:.1f}, " +\
                " $\\Delta^{\\ddag} E_{el}^{b}$: " + f" {barrier_backward:.1f}"
        if from_spline:
            label += "\n$^{*}$ Energies from Spline"

        self.ax1.set_xlabel(label)

    def draw_point(self, current_frame: int, total_frames: int):
        if self._prev_point is not None:
            self._prev_point.remove()
        if self._spline is None:
            return
        x = current_frame / total_frames
        if self._invert:
            x = 1.0 - x
        if x < 1e-6:
            y = 0.0
        else:
            y_base = self._spline.evaluate(0, self._spline_order)[0] if not self._invert else\
                self._spline.evaluate(1, self._spline_order)[0]
            y = (self._spline.evaluate(x, self._spline_order)[0] - y_base) * utils.KJPERMOL_PER_HARTREE
        self._prev_point = self.ax1.scatter(current_frame / total_frames, y,
                                            color=get_primary_light_color(), zorder=2.5)
        self.canvas.draw()

    def save_svg(self):
        filename, _ = QFileDialog.getSaveFileName(
            self,
            self.tr("Save File"),  # type: ignore[arg-type]
            "plot.svg",
            self.tr("Vector Graphics (*.svg)"),  # type: ignore[arg-type]
        )
        if self._prev_point is not None:
            try:
                self._prev_point.remove()
                self._prev_point = None
            except BaseException:
                pass
        self.canvas.draw()
        self.fig.savefig(filename, transparent=True)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""

from PySide2.QtWidgets import QSpinBox, QDoubleSpinBox

from scine_heron.containers.combo_box import BaseBox


class NoWheelSpinBox(QSpinBox):
    """
    QSpinBox that does not react to mouse wheel events
    """

    def wheelEvent(self, event) -> None:
        event.ignore()


class NoWheelDoubleSpinBox(QDoubleSpinBox):
    """
    QDoubleSpinBox that does not react to mouse wheel events
    """

    def wheelEvent(self, event) -> None:
        event.ignore()


class NoWheelComboBox(BaseBox):
    """
    QComboBox that does not react to mouse wheel events
    """

    def wheelEvent(self, event) -> None:
        event.ignore()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""

from typing import Optional
import scine_database as db

from PySide2.QtCore import QObject
from scine_heron.containers.combo_box import BaseBox


class SideBuilder(BaseBox):
    """
    Selection for the database Side enum.
    """

    def __init__(self, parent: Optional[QObject] = None):
        super().__init__(parent=parent)
        self.addItems(["BOTH", "LHS", "RHS"])
        self.sides = {
            "BOTH": db.Side.BOTH,
            "LHS": db.Side.LHS,
            "RHS": db.Side.RHS,
        }

    def get_side(self) -> db.Side:
        assert self.currentText()
        return self.sides[self.currentText()]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""

from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional

import scine_utilities as su
import numpy as np


@dataclass
class CustomResult:
    result: su.Results = field(default_factory=su.Results)
    energy: float = field(default=0.0, init=False)
    gradients: Optional[np.ndarray] = field(default=None, init=False)
    hessian: np.ndarray = field(init=False)
    positions: Optional[np.ndarray] = field(default=None)
    atomic_charges: Optional[List[float]] = field(init=False)
    bond_orders: Optional[np.ndarray] = field(init=False)
    molden_input: str = field(default="")
    error_msg: str = field(default="")
    info_msg: str = field(default="")
    settings: Dict[str, Any] = field(default_factory=dict)

    def __post_init__(self):
        if self.result.successful_calculation is None:
            self.successful = False
            return self
        if not self.result.successful_calculation:
            return CustomResult()
        self.energy: float = self.result.energy
        self.gradients: np.ndarray = self.result.gradients
        self.hessian: np.ndarray = self._construct_hessian()
        self.atomic_charges: List[float] = self.result.atomic_charges
        self.bond_orders: np.ndarray = self._construct_bond_matrix()
        self.successful: bool = self.result.successful_calculation is not None and self.result.successful_calculation

    def _construct_hessian(self) -> Optional[np.ndarray]:
        if self.result.atomic_hessian is not None and self.result.atomic_charges is not None:
            size = len(self.result.atomic_charges)
            # Convert list of atomic hessians to block-diagonal pos. definite approx. hessian
            atomic_hessians = self.result.atomic_hessian.get_atomic_hessians()
            new_atomic_hessians = []
            for atomic_hessian in atomic_hessians:
                eigvals, eigvecs = np.linalg.eig(atomic_hessian)
                new_eigvals = abs(eigvals)
                new_atomic_hessians.append(eigvecs @ np.diag(new_eigvals) @ eigvecs.T)
            new_hessian = np.zeros((3 * size, 3 * size,))
            for i in range(size):
                new_hessian[i * 3: i * 3 + 3, i * 3: i * 3 + 3] = new_atomic_hessians[i]
        elif self.result.hessian is not None:
            hessian = self.result.hessian
            # Make hessian pos. definite
            eigvals, eigvecs = np.linalg.eig(hessian)
            new_eigvals = abs(eigvals)
            new_hessian = eigvecs @ np.diag(new_eigvals) @ eigvecs.T
        else:
            new_hessian = None
        return new_hessian

    def _construct_bond_matrix(self) -> Optional[np.ndarray]:
        if self.result.bond_orders is not None:
            return self.result.bond_orders.matrix.toarray()
        return None

    def __bool__(self):
        return self.successful

    def make_pickleable(self):
        self.result = None



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""
"""
Provides the MoleculeWidget class.
"""
from PySide2.QtWidgets import (
    QWidget,
    QVBoxLayout,
    QSlider,
    QFileDialog,
    QMainWindow,
    QDialog,
    QApplication,
)
from scine_heron.utilities import (
    color_axis,
    color_figure,
    get_font,
    get_primary_line_color,
    get_secondary_line_color,
)
from scine_heron.molecule.molecule_widget import MoleculeWidget
from scine_heron.toolbar.io_toolbar import HeronToolBar
from PySide2.QtCore import QObject, Qt

from matplotlib.collections import PathCollection
from matplotlib.figure import Figure
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg
from matplotlib.text import Annotation
from matplotlib.widgets import RectangleSelector


from scipy.signal import argrelextrema
import numpy as np
from pkgutil import iter_modules

import scine_utilities as utils
from typing import Any, Tuple, Optional, List, Set, TYPE_CHECKING
if TYPE_CHECKING:
    Signal = Any
else:
    from PySide2.QtCore import Signal


class MoleculeVideo(QWidget):
    """
    Displays trajectory as interactive video
    """

    changed_frame = Signal(int, int)

    def __init__(
        self,
        parent: Optional[QObject],
        trajectory: utils.MolecularTrajectory,
        mol_widget: MoleculeWidget
    ):
        QWidget.__init__(self, parent)
        if trajectory.empty():
            raise RuntimeError("Empty Trajectory given to video")
        self._frame: int = 0
        self._n_frames: int = len(trajectory)
        self._trajectory: utils.MolecularTrajectory = trajectory
        self._mol_widget = mol_widget
        self._mol_widget.update_molecule(atoms=utils.AtomCollection(
            trajectory.elements, trajectory[0]))
        layout = QVBoxLayout()
        # By adding the mol widget to this layout it is losing its original layout info
        layout.addWidget(self._mol_widget)

        self._tool_bar: Optional[HeronToolBar] = None
        self.slider: Optional[QSlider] = None
        if self._n_frames > 1:
            self._tool_bar = HeronToolBar(parent=self)

            self.slider = QSlider(Qt.Horizontal)
            self.slider.setMinimum(0)
            self.slider.setMaximum(self._n_frames - 1)
            self.slider.setValue(0)
            self.slider.valueChanged.connect(self.slider_change)  # pylint: disable=no-member
            self._tool_bar.addWidget(self.slider)

            self._tool_bar.shortened_add_action("save_molecule.png", "Save frame", "",
                                                self._save_frame)
            self._tool_bar.addSeparator()
            self._tool_bar.shortened_add_action('save_trajectory.png', "Save trajectory", "",
                                                self._save_trajectory)

            layout.addWidget(self._tool_bar)

        self.setLayout(layout)

    def add_button_to_toolbar(self, *args, **kwargs):
        if self._tool_bar:
            self._tool_bar.shortened_add_action(*args, **kwargs)

    def slider_change(self) -> None:
        if self.slider is not None:
            self._frame = self.slider.value()
            self._mol_widget.update_molecule(atoms=utils.AtomCollection(
                self._trajectory.elements, self._trajectory[self._frame]), reset_camera=False)
            self.changed_frame.emit(self._frame, self._n_frames)

    def set_frame(self, frame: int) -> None:
        if self.slider is not None:
            if frame < 0 or frame >= self._n_frames:
                frame = 0
            self.slider.setValue(frame)

    def _save_trajectory(self) -> None:
        filename, _ = QFileDialog.getSaveFileName(
            self,
            self.tr("Save Trajectory to File"),  # type: ignore[arg-type]
            "",
            self.tr("Molecule (*.xyz *.bin)"),  # type: ignore[arg-type]
        )
        if filename:
            if str(filename)[-3:].lower() == "xyz":
                traj_format = utils.io.TrajectoryFormat.Xyz
            elif str(filename)[-3:].lower() == "bin":
                traj_format = utils.io.TrajectoryFormat.Binary
            else:
                raise RuntimeError(f"Unknown trajectory format given in {str(filename)}, "
                                   f"please specify either 'xyz' or 'bin'.")
            utils.io.write_trajectory(traj_format, str(filename), self._trajectory)

    def _save_frame(self) -> None:
        filename, _ = QFileDialog.getSaveFileName(
            self,
            self.tr("Save Frame to File"),  # type: ignore[arg-type]
            "",
            self.tr("Molecule (*.xyz *.mol *.pdb)"),  # type: ignore[arg-type]
        )
        if filename:
            atoms = utils.AtomCollection(self._trajectory.elements, self._trajectory[self._frame])
            utils.io.write(str(filename), atoms)


class TrajectoryEnergyWidget(FigureCanvasQTAgg):
    def __init__(
            self, parent: QWidget, trajectory: utils.MolecularTrajectory,
            width: float = 20, height: float = 3, selectable: bool = True
    ):
        self.fig = Figure(figsize=(width, height))
        color_figure(self.fig)
        self.ax = self.fig.add_subplot(1, 1, 1)
        color_axis(self.ax)

        super(TrajectoryEnergyWidget, self).__init__(self.fig)
        self.__selectable: bool = selectable
        self._energies = trajectory.get_energies()
        if self._energies:
            self.start = 1.0
            for e in self._energies:
                if e != float('inf'):
                    self.start = e
                    break
            self.end = 1.0
            for e in reversed(self._energies):
                if e != float('inf'):
                    self.end = e
                    break
            self.x = list(range(len(self._energies)))
            for i in reversed(range(len(self._energies))):
                if self._energies[i] == float('inf'):
                    self.x.pop(i)
            self.from_frame_map: List[Optional[int]] = [None for _ in range(len(self._energies))]
            counter = 0
            for i in range(len(self._energies)):
                if self._energies[i] != float('inf'):
                    self.from_frame_map[i] = counter
                    counter += 1
            self.y = [(self._energies[i] - self.start) * 2625.5 for i in self.x]
            self.y_plus_interpolation = []
            self.y_interpolation = []
            self.x_interpolation = []
            current_x = 0
            if self.from_frame_map[0] is None:
                for current_x, val in enumerate(self.from_frame_map):
                    if val is not None:
                        break
                    if len(self.y) > 0:
                        self.y_plus_interpolation.append(self.y[0])
                        self.y_interpolation.append(self.y[0])
                    else:
                        self.y_plus_interpolation.append(0.0)
                        self.y_interpolation.append(0.0)
                    self.x_interpolation.append(current_x)
            for i in range(len(self.x) - 1):
                if self.x[i + 1] != self.x[i] + 1:
                    stride = (self.x[i + 1] - self.x[i])
                    diff = (self.y[i + 1] - self.y[i])
                    for j in range(stride):
                        value = self.y[i] + j * diff / stride
                        self.y_plus_interpolation.append(value)
                        self.y_interpolation.append(value)
                        self.x_interpolation.append(current_x)
                        current_x += 1
                else:
                    self.y_plus_interpolation.append(self.y[i])
                    current_x += 1
            if len(self.y) > 0:
                self.y_plus_interpolation.append(self.y[-1])
            else:
                self.y_plus_interpolation.append(0.0)
            current_x += 1
            for x in range(current_x, len(self._energies)):
                if len(self.y) > 0:
                    self.y_plus_interpolation.append(self.y[-1])
                    self.y_interpolation.append(self.y[-1])
                else:
                    self.y_plus_interpolation.append(0.0)
                    self.y_interpolation.append(0.0)
                self.x_interpolation.append(x)
            self._extrema: np.ndarray = np.array([])
            self._prev_point = None
            self._parent = parent
            self._scatters: Optional[PathCollection] = None
            self._annotations: Optional[Annotation] = None
            self._mark_color = "#bc80bd"
        self.selected_points: Set[int] = set()
        self.update_canvas()

    def update_canvas(self):
        if not self._energies:
            self.fig.set_visible(False)
            self.draw()
            return
        if len(self.y) > 0:
            rel_maximum = np.max(self.y)
            rel_max_pos = self.x[np.argmax(self.y)]
        else:
            rel_maximum = 0.0
            rel_max_pos = 0
        de = (self.end - self.start) * 2625.5
        dedf = rel_maximum
        dedb = rel_maximum + (self.start - self.end) * 2625.5
        self.ax.cla()
        self.ax.plot(self.x_interpolation, self.y_interpolation,
                     color=get_primary_line_color(), linestyle='', marker='x', markersize=5,
                     picker=True, pickradius=5)
        self.ax.plot(self.x, self.y,
                     color=get_primary_line_color(), linestyle='', marker='o', markersize=5,
                     picker=True, pickradius=5)
        start_x = 0
        for i in range(len(self.x) - 1):
            if self.x[i + 1] != self.x[i] + 1:
                self.ax.plot(self.x[start_x:i + 1], self.y[start_x:i + 1],
                             color=get_primary_line_color(), linestyle='--')
                start_x = i + 1
            elif i + 2 == len(self.x):
                self.ax.plot(self.x[start_x:i + 2], self.y[start_x:i + 2],
                             color=get_primary_line_color(), linestyle='--')
        font = get_font()
        self.ax.set_title(
            f"Maximum energy at ({rel_max_pos}, {rel_maximum:.1f})",
            font, color=get_primary_line_color()
        )
        self.ax.set_xlabel(
            f"dE: {de:.1f}, Ef: {dedf:.1f}, Eb: {dedb:.1f}", font
        )
        self.ax.set_ylabel("Relative energy in kJ/mol", font)
        self.fig.tight_layout()
        if self._parent is not None and isinstance(self._parent, MoleculeVideoWithEnergy):
            self._parent.video.changed_frame.connect(self.draw_point)
        self.fig.canvas.mpl_connect('pick_event', self._onpick)
        self.fig.canvas.mpl_connect('motion_notify_event', self._hover)
        self._add_extrema_infos()
        self.draw_point(0, 1)

        # Handle selection of trajectory points
        self.selected_points = set()
        if self.__selectable:
            self.selection, = self.ax.plot([], [],
                                           color=get_secondary_line_color(), linestyle='', marker='o', markersize=10,
                                           picker=True, pickradius=5, zorder=0)

            def toggle_selector(event):
                print(' Key pressed.')
                if event.key in ['Q', 'q'] and toggle_selector.RS.active:
                    print(' RectangleSelector deactivated.')
                    toggle_selector.RS.set_active(False)
                if event.key in ['A', 'a'] and not toggle_selector.RS.active:
                    print(' RectangleSelector activated.')
                    toggle_selector.RS.set_active(True)
            toggle_selector.RS = RectangleSelector(self.ax, self.line_select_callback,
                                                   drawtype='box', useblit=True,
                                                   button=[1, 3],  # don't use middle button
                                                   minspanx=5, minspany=5,
                                                   spancoords='pixels',
                                                   interactive=False)
            self.fig.canvas.mpl_connect('key_press_event', toggle_selector)

    def line_select_callback(self, eclick, erelease):
        'eclick and erelease are the press and release events'
        x1, y1 = eclick.xdata, eclick.ydata
        x2, y2 = erelease.xdata, erelease.ydata
        newly_selected_points: Set[int] = set()
        for x, y in zip(range(len(self._energies)), self.y_plus_interpolation):
            if not (x1 > x > x2 or x2 > x > x1):
                continue
            if not (y1 > y > y2 or y2 > y > y1):
                continue
            newly_selected_points.add(x)
        if eclick.button == 1:
            self.selected_points.update(newly_selected_points)
        elif eclick.button == 3:
            self.selected_points = self.selected_points - newly_selected_points
        self.update_selection()

    def update_selection(self):
        x = list(self.selected_points)
        self.selection.set_xdata(x)
        self.selection.set_ydata([self.y_plus_interpolation[i] for i in x])
        self.fig.canvas.draw()
        self.fig.canvas.flush_events()

    def save_svg(self):
        filename, _ = QFileDialog.getSaveFileName(
            self,
            self.tr("Save File"),  # type: ignore[arg-type]
            "plot.svg",
            self.tr("Vector Graphics (*.svg)"),  # type: ignore[arg-type]
        )
        self.fig.savefig(filename)

    def draw_point(self, current_frame: int, _: int):
        if self._prev_point is not None:
            self._prev_point.remove()
            self._prev_point = None
        if not self._energies:
            return
        current_y = self.y_plus_interpolation[current_frame]
        self._prev_point = self.ax.scatter(current_frame, current_y, color="C1", zorder=100)
        self.draw()

    def _find_local_minima_and_maxima(self, n_local: int = 5) -> Tuple[np.ndarray, np.ndarray]:
        values = np.array(self.y_plus_interpolation)
        maxima = argrelextrema(values, np.greater, order=n_local)[0]
        minima = argrelextrema(values, np.less, order=n_local)[0]
        extrema = np.asarray(list(maxima) + list(minima))
        extreme_values = np.array([values[x] for x in extrema])
        self._extrema = extrema
        return self._extrema, extreme_values

    def _add_extrema_infos(self) -> None:
        sc_x, sc_y = self._find_local_minima_and_maxima()
        sc = self.ax.scatter(sc_x, sc_y, color=self._mark_color, zorder=99)
        self._scatters = sc
        annot = self.ax.annotate("", xy=(0, 0), xytext=(20, 20), textcoords="offset points",
                                 bbox=dict(boxstyle="round", fc="w"), arrowprops=dict(arrowstyle="->"))
        annot.set_visible(False)
        self._annotations = annot

    def _update_annot(self, index) -> None:
        if self._scatters is not None and self._annotations is not None:
            index = index['ind'][0]
            text = "rel. energy: {0:.2f}".format(self.y_plus_interpolation[self._extrema[index]])
            annot = self._annotations
            annot.xy = self._scatters.get_offsets()[index]
            annot.set_text(text)
            annot.get_bbox_patch().set_facecolor(self._mark_color)
            annot.get_bbox_patch().set_alpha(0.4)

    def _hover(self, event) -> None:
        if event.inaxes == self.ax and self._scatters is not None and self._annotations is not None:
            contains, index = self._scatters.contains(event)
            annot = self._annotations
            visible = annot.get_visible()
            if contains:
                self._update_annot(index)
                annot.set_visible(True)
                self.fig.canvas.draw_idle()
            else:
                if visible:
                    annot.set_visible(False)
                    self.fig.canvas.draw_idle()

    def _onpick(self, event) -> None:
        if self._annotations is not None:
            self._annotations.set_visible(False)
            if self._parent is not None and isinstance(self._parent, MoleculeVideoWithEnergy):
                x = int(round(event.mouseevent.xdata))
                self._parent.video.set_frame(x)


def find_main_window() -> Optional[QMainWindow]:
    # Global function to find the (open) QMainWindow in application
    # Copy of function in __init__.py as this file is imported there
    app = QApplication.instance()
    for widget in app.topLevelWidgets():
        if isinstance(widget, QMainWindow):
            return widget
    return None


class MoleculeVideoWithEnergy(QWidget):

    def __init__(
            self, parent: QWidget, trajectory: utils.MolecularTrajectory,
            width: float = 20, height: float = 3, alternative_zoom_controls: bool = False,
            selectable: bool = True
    ):
        QWidget.__init__(self, parent)
        self.molecule_widget = MoleculeWidget(
            parent=self,
            alternative_zoom_controls=alternative_zoom_controls,
            disable_modification=True
        )
        self.__trajectory = trajectory
        self.video = MoleculeVideo(parent=self, trajectory=trajectory, mol_widget=self.molecule_widget)
        self.plot = TrajectoryEnergyWidget(parent=self, trajectory=trajectory,
                                           width=width, height=height, selectable=selectable)

        self.plot_tool_bar = HeronToolBar(parent=self)
        self.plot_tool_bar.shortened_add_action('save_plot.png', "Save plot", "",
                                                self.plot.save_svg)
        if selectable:
            self.plot_tool_bar.shortened_add_action('save_trajectory.png', "Save Selected Trajectory", "",
                                                    self.__save_selected_trajectory)
            self.plot_tool_bar.shortened_add_action('save_trajectory.png', "Preview Selected Trajectory", "",
                                                    self.__preview_trajectory)

            # Check if there is Scine Art and something to store templates in
            main_window = find_main_window()
            has_rt_storage = False
            if main_window is not None and hasattr(main_window, 'get_reaction_template_storage'):
                rt_storage = main_window.get_reaction_template_storage()
                if rt_storage is not None:
                    has_rt_storage = True

            if "scine_art" in (name for _, name, _ in iter_modules()) and has_rt_storage:
                # If available allow for template extraction from the trajectory
                self.plot_tool_bar.shortened_add_action(
                    'save_trajectory.png',
                    'Extract Reaction Template from Selected Trajectory',
                    '',
                    self.__extract_template
                )

        layout = QVBoxLayout()
        layout.addWidget(self.video)
        layout.addWidget(self.plot)
        layout.addWidget(self.plot_tool_bar)
        self.setLayout(layout)

    def __save_selected_trajectory(self):
        selected_frames = list(self.plot.selected_points)
        if not selected_frames:
            trajectory = self.__trajectory
        else:
            trajectory = utils.MolecularTrajectory(self.__trajectory.elements)
            for i in selected_frames:
                trajectory.push_back(self.__trajectory[i], self.__trajectory.get_energies()[i])

        filename, _ = QFileDialog.getSaveFileName(
            self,
            self.tr("Save Trajectory to File"),  # type: ignore[arg-type]
            "",
            self.tr("Molecule (*.xyz *.bin)"),  # type: ignore[arg-type]
        )
        if filename:
            if str(filename)[-3:].lower() == "xyz":
                traj_format = utils.io.TrajectoryFormat.Xyz
            elif str(filename)[-3:].lower() == "bin":
                traj_format = utils.io.TrajectoryFormat.Binary
            else:
                raise RuntimeError(f"Unknown trajectory format given in {str(filename)}, "
                                   f"please specify either 'xyz' or 'bin'.")
            utils.io.write_trajectory(traj_format, str(filename), trajectory)

    def __preview_trajectory(self):
        selected_frames = list(self.plot.selected_points)
        if not selected_frames:
            MoleculeVideoWithEnergyDialog(self, self.__trajectory, selectable=False)
        else:
            selected_trajectory = utils.MolecularTrajectory(self.__trajectory.elements)
            for i in selected_frames:
                selected_trajectory.push_back(self.__trajectory[i], self.__trajectory.get_energies()[i])
            MoleculeVideoWithEnergyDialog(self, selected_trajectory, selectable=False)

    def __extract_template(self):
        selected_frames = list(self.plot.selected_points)
        if not selected_frames:
            trajectory = self.__trajectory
        else:
            trajectory = utils.MolecularTrajectory(self.__trajectory.elements)
            for i in selected_frames:
                trajectory.push_back(self.__trajectory[i], self.__trajectory.get_energies()[i])
        from scine_art.reaction_template import ReactionTemplate
        main_window = find_main_window()
        rt_storage = main_window.get_reaction_template_storage()
        reaction_template = ReactionTemplate.from_trajectory(trajectory)
        rt_storage.add_reaction_template(reaction_template)


class MoleculeVideoWithEnergyDialog(QDialog):
    def __init__(
            self, parent: QWidget, trajectory: utils.MolecularTrajectory,
            width: float = 20, height: float = 3, alternative_zoom_controls: bool = False,
            selectable: bool = True, window_title: str = "Trajectory History",
    ) -> None:
        super(MoleculeVideoWithEnergyDialog, self).__init__(parent)
        self.setWindowTitle(window_title)

        # Create layout and add widgets
        layout = QVBoxLayout()
        self.video = MoleculeVideoWithEnergy(self, trajectory, width, height,
                                             alternative_zoom_controls, selectable)
        layout.addWidget(self.video)

        # Set dialog layout
        self.setLayout(layout)
        self.exec_()


class MainVideo(QMainWindow):
    def __init__(
            self,
            trajectory,
            parent: Optional[QObject] = None,
    ):
        QMainWindow.__init__(self, parent)
        self.resize(1280, 1024)

        self.setWindowTitle(self.tr("Trajectory"))  # type: ignore[arg-type]
        self.setCentralWidget(MoleculeVideoWithEnergy(self, trajectory))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""

from typing import List, Optional

import numpy as np
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg
from matplotlib.figure import Figure
from PySide2.QtWidgets import QDockWidget, QGridLayout, QWidget
from scine_autocas.plots.entanglement_plot import EntanglementPlot

from scine_heron.autocas.signal_handler import SignalHandler


class EntanglementDiagramWidget(QDockWidget):
    class Canvas(FigureCanvasQTAgg):
        def __init__(
            self,
            # pylint: disable-next=W0613
            mode: str = "dark",
            # pylint: disable-next=W0613
            parent: Optional[QWidget] = None,
            width: float = 10,
            height: float = 10,
        ):

            self.entang_plot = EntanglementPlot()
            self.entang_plot.alpha_mut_inf = 1
            self.entang_plot.alpha_s1 = 1
            self.entang_plot.s1_position = 1.0
            # self.entang_plot.label_ofset = 0.35

            self.fig = Figure(figsize=(width, height))
            self.ax1 = self.fig.add_subplot(
                111, polar=True, position=[0.05, 0.05, 0.9, 0.9]
            )
            super(EntanglementDiagramWidget.Canvas, self).__init__(self.fig)

            self.s1 = np.array([1.0, 1.0])
            self.mutual_information = np.array([[1.0, 1.0], [1.0, 1.0]])
            self.order: List[int] = [0, 1]

            # self.__lines = self.entang_plot._entanglement_plot(
            #     self.ax1, self.s1, self.mutual_information, self.order
            # )
            self.entang_plot._entanglement_plot(
                self.ax1, self.s1, self.mutual_information, self.order
            )

            self.draw()

        def update_lines(self, s1=None, mutual_information=None, order=None):
            if s1 is None:
                s1 = np.array([0.0, 0.0])
            if mutual_information is None:
                mutual_information = np.array([[0.0, 0.0], [0.0, 0.0]])

            self.s1 = s1
            self.mutual_information = mutual_information
            self.order = order
            # self.entang_plot._entanglement_plot(
            #    self.ax1, self.s1, self.mutual_information, None
            # )
            # self.__line = self.entang_plot._entanglement_plot(
            #     self.ax1, self.s1, self.mutual_information, self.order
            # )
            self.entang_plot._entanglement_plot(
                self.ax1, self.s1, self.mutual_information, self.order
            )
            # local_x = self.x
            # local_y = self.y
            # x = np.array(local_x)
            # y = np.array(local_y)
            # x_min = np.min(x)
            # x_max = np.max(x)
            # y_min = np.min(y)
            # y_max = np.max(y)
            # self.__line.set_data(x, y)
            # self.__line.axes.axis([x_min, x_max, y_min, y_max])

            self.draw()

    def __init__(
        self,
        # pylint: disable-next=W0613
        parent: Optional[QWidget],
        signal_handler: SignalHandler,
        # pylint: disable-next=W0613
        mode: str = "dark",
        width: float = 10,
        height: float = 10,
    ) -> None:
        super(EntanglementDiagramWidget, self).__init__()
        self.setFeatures(QDockWidget.NoDockWidgetFeatures)

        self.signal_handler = signal_handler
        self.signal_handler.update_entanglement_plot_signal.connect(self.update_plot)

        self._inner_widget = QWidget(self)
        self.setWidget(self._inner_widget)
        self._layout = QGridLayout()
        self._layout.setRowMinimumHeight(0, 200)
        self._layout.setColumnMinimumWidth(0, 200)
        self._inner_widget.setLayout(self._layout)
        # self.__widget_width = 130
        # self.__widget_height = 130
        self.__canvas = self.Canvas(parent=self, width=width, height=height, mode=mode)
        self._layout.addWidget(self.__canvas, 0, 0, 1, 1)
        self.__canvas.update_lines()

    def update_plot(self, s1, mut_inf, order=None):
        self.__canvas.ax1.cla()
        self.__canvas.update_lines(s1, mut_inf, order)
        self.__canvas.draw()

    def plot_entanglement(self):
        pass

    def plot_threshold(self):
        pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""
"""
Provides the MoleculeDepthView class
and its subclasses.
"""
from vtkmodules.qt.QVTKRenderWindowInteractor import QVTKRenderWindowInteractor
from vtk import (
    vtkMoleculeMapper,  #
    vtkMolecule,  #
    vtkActor,  #
    vtkRenderer,  #
    vtkRenderWindow,  #
    vtkInteractorStyle,  #
    vtkAlgorithmOutput,  #
    vtkPolyDataMapper,  #
    vtkSphereSource,
)  #

import scine_heron.config as config
from scine_heron.utilities import hex_to_rgb_base_1

from PySide2.QtWidgets import QWidget
from scine_heron.molecule.depth_view_algorithms import DepthProjection

from abc import ABC, abstractmethod
from typing import Any


class MoleculeDepthView(ABC):
    """
    A base class that wraps a QVTKRenderWindowInteractors
    displaying an additional representation of a molecule
    and a haptic pointer.
    It exposes:
    - the interactor, so that it can be added
      to existing QWidgets and QLayouts;
    - a callback for camera change,
      that can be used as an observer
      for the RenderEvent of a vtkRenderWindow
      (i.e., the main renderwindow,
      where the main view of the molecule
      is displayed.)
    - an input port for a molecule
    - an input port for the haptic pointer data
    It is responsibility of the owner of this object
    to pass the `camera_changed` callback to the right caller
    (i.e., to use the `AddObserver` method
    on the main RenderWindow).
    """

    def __init__(self, parent: QWidget) -> None:
        # Haptic pointer pipeline - general
        # This needs to be connected with the subclass-specific pipeline
        self._haptic_pointer_mapper = vtkPolyDataMapper()
        self._haptic_pointer_actor = vtkActor()
        self._haptic_pointer_actor.SetMapper(self._haptic_pointer_mapper)
        self._haptic_pointer_actor.GetProperty().SetColor(0, 0.75, 0.75)

        # Molecule vis graphic pipeline - general
        self._molecule_mapper = vtkMoleculeMapper()
        self._molecule_actor = vtkActor()
        self._molecule_actor.SetMapper(self._molecule_mapper)

        # renderer
        self._renderer = vtkRenderer()
        self._renderer.SetPreserveColorBuffer(True)
        self._renderer.SetLayer(0)
        rgb = hex_to_rgb_base_1(config.COLORS['secondaryLightColor'])
        self._renderer.SetBackground(*rgb)
        self._renderer.AddActor(self._molecule_actor)
        self._renderer.AddActor(self._haptic_pointer_actor)

        # interactor
        self._interactor = QVTKRenderWindowInteractor(parent=parent)
        self._interactor.Initialize()
        self._interactor.GetRenderWindow().AddRenderer(self._renderer)

        # interactor style, which does nothing
        self._style = vtkInteractorStyle()
        self._interactor.SetInteractorStyle(self._style)

    @abstractmethod
    def set_molecule_input(self, output_port: vtkAlgorithmOutput) -> None:
        """
        This method takes the output ports of an algorithm
        producing a vtkMolecule
        and connects it to the part of the molecule visualization pipeline
        for which this object is responsible.
        """

    @abstractmethod
    def set_haptic_pointer_data_input(self, output_port: vtkAlgorithmOutput) -> None:
        """
        This method takes the output port of an algorithm
        producing a (x,y,z) vtkPolyData
        and connects it to the haptic pointer visualization pipeline
        in this object.
        """

    @property
    def interactor(self) -> QVTKRenderWindowInteractor:
        return self._interactor

    @abstractmethod
    def camera_changed(self, main_window: vtkRenderWindow, _2: Any) -> None:
        """
        This method signature allows it to be used
        as an Observer for a vtkRenderWindow
        for a RenderEvent.
        """

    def render(self) -> None:
        self._renderer.GetRenderWindow().Render()

    def set_molecule(self, molecule: vtkMolecule) -> None:
        """
        Sets the molecule object that is represented
        """


class DepthViewZ1dFixedOnHapticSoftMaxMin(MoleculeDepthView):
    def __init__(self, parent: QWidget) -> None:
        super(DepthViewZ1dFixedOnHapticSoftMaxMin, self).__init__(parent)

        # The maximum distance from the haptic pointer
        # beyond which the atoms will not be plotted
        # in the depth view.
        self.filter_range = 5

        # The "saturation" distance of the atoms
        # from the haptic pointer (i.e., the origin)
        # in the depth view
        self.zscale = 3
        self._projection = DepthProjection(
            filter_range=self.filter_range, zscale=self.zscale
        )
        self._molecule_mapper.SetInputConnection(self._projection.output)
        self._molecule_mapper.SetAtomicRadiusScaleFactor(0.15)  # default is 0.3

        # The haptic pointer in this view is a sphere always in the center.
        # NOTE: we can move it if needed,
        #       e.g. to the left of the right
        #       to improve readability.
        self._haptic_pointer_source = vtkSphereSource()
        self._haptic_pointer_source.SetCenter(0, 0, 0)
        self._haptic_pointer_source.SetRadius(0.25)
        self._haptic_pointer_source.SetPhiResolution(25)
        self._haptic_pointer_source.SetThetaResolution(25)

        # Hooking the haptic data
        self._haptic_pointer_mapper.SetInputConnection(
            self._haptic_pointer_source.GetOutputPort()
        )

    def set_molecule_input(self, output_port: vtkAlgorithmOutput) -> None:
        """
        We connect the molecule output
        and connect it to the input of the depth projection algorithm.
        """
        self._projection.set_molecule_input(output_port)

    def set_haptic_pointer_data_input(self, output_port: vtkAlgorithmOutput) -> None:
        """
        Set the polydata connection for the haptic data.
        It must be fed directly to the projection algorithm.
        """
        self._projection.set_haptic_pointer_data(output_port)

    def camera_changed(self, main_window: vtkRenderWindow, _2: Any) -> None:
        main_camera = main_window.GetRenderers().GetFirstRenderer().GetActiveCamera()
        main_depth_direction = main_camera.GetDirectionOfProjection()

        # Camera orientation is the same - it's the atoms in the depth view that move.
        # But we change the zoom and the distance
        camera = self._renderer.GetActiveCamera()
        camera.SetFocalPoint((0, 0, 0))
        camera.SetViewUp((0, 0, 1))

        # to make sure no atom is cut in half in the depth view
        distance_safety_margin = 1.15
        distance = self.zscale * distance_safety_margin
        camera.SetPosition((0, distance, 0))
        camera.SetViewAngle(90)  # +- 45 to make calculations easy.
        camera.SetClippingRange([0.001, distance + 5])

        self._projection.set_camera_data(main_depth_direction)
        self.render()
        # There is no haptic pointer projection to update.

    def set_molecule(self, molecule: vtkMolecule) -> None:
        self._projection.set_molecule(molecule)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""
"""
This module provides two functions of jsonpickle with hardcoded defaults and
it ensures that we serialize all our own classes correctly as soon as this module is loaded.
jsonpickle allows to safe an object in an human-readable format.
"""


from enum import Enum
from typing import Any, TYPE_CHECKING
import jsonpickle
import jsonpickle.ext.numpy as jsonpickle_numpy

from scine_utilities import ElementInfo, ElementType, AtomCollection

from scine_heron.dependencies.optional_import import importer, is_imported

if TYPE_CHECKING:
    from scine_chemoton.steering_wheel.datastructures import Status, LogicCoupling
    from scine_chemoton.utilities.reactive_complexes.lebedev_sphere import LebedevSphere
    from scine_database import Label
    from scine_chemoton.utilities.place_holder_model import PlaceHolderModelType, _PlaceHolderModelProxy
else:
    Status = importer("scine_chemoton.steering_wheel.datastructures", "Status")
    LogicCoupling = importer("scine_chemoton.steering_wheel.datastructures", "LogicCoupling")
    LebedevSphere = importer("scine_chemoton.utilities.reactive_complexes.lebedev_sphere", "LebedevSphere")
    Label = importer("scine_database", "Label")
    PlaceHolderModelType = importer("scine_chemoton.utilities.place_holder_model", "PlaceHolderModelType")
    _PlaceHolderModelProxy = importer("scine_chemoton.utilities.place_holder_model", "_PlaceHolderModelProxy")


class EnumHandler(jsonpickle.handlers.BaseHandler):
    """
    This handler exists because jsonpickle has problems with our Enum derived classes.
    """

    def flatten(self, obj, data):
        data['value'] = self.context.flatten(repr(obj), reset=False)
        return data

    def restore(self, obj):
        return self.context.restore(eval(obj['value'].split("<")[-1].split(":")[0]),  # pylint: disable=eval-used
                                    reset=False)


class ProxyHandler(jsonpickle.handlers.BaseHandler):
    """
    This class circumvents jsonpickle's problem with Proxy objects
    """
    _proxy = _PlaceHolderModelProxy

    def flatten(self, obj, data):
        data['state'] = self.context.flatten(obj.__getstate__(), reset=False)
        return data

    def restore(self, obj):
        proxy = self._proxy()
        proxy.__setstate__(obj['state'])
        return self.context.restore(proxy,
                                    reset=False)


class SimpleReprEvalHandler(jsonpickle.handlers.BaseHandler):
    """
    This handler offers an easy implementation for all classes that work with an
    eval(repr(inst)) call.
    """

    def flatten(self, obj, data):
        data['value'] = self.context.flatten(repr(obj), reset=False)
        return data

    def restore(self, obj):
        return self.context.restore(eval(obj['value']),  # pylint: disable=eval-used
                                    reset=False)


class ElementHandler(jsonpickle.handlers.BaseHandler):
    """
    This handler stores Scine Utilities ElementTypes by their string representation.
    To be removed once pickling support is available from Scine Utilities.
    """

    def flatten(self, obj, data):
        data['value'] = self.context.flatten(str(obj), reset=False)
        return data

    def restore(self, obj):
        return self.context.restore(ElementInfo.element_from_symbol(obj['value']), reset=False)


class AtomCollectionHandler(jsonpickle_numpy.NumpyGenericHandler):
    """
    This handler stores Scine AtomCollections by saving elements and positions separately.
    It inherits from NumpyGenericHandler to store the positions as numpy arrays.
    """

    def flatten(self, obj, data):
        pos_data = {}
        data['positions'] = super().flatten(obj.positions, pos_data)
        data['elements'] = self.context.flatten(repr([str(e) for e in obj.elements]), reset=False)
        return data

    def restore(self, data):
        elements = self.context.restore([ElementInfo.element_from_symbol(e)
                                         for e in eval(data['elements'])],  # pylint: disable=eval-used
                                        reset=False)
        positions = super().restore(data['positions'])
        return AtomCollection(elements, positions)


def encode(obj: Any) -> str:
    """
    A wrap with default parameters for jsonpickle.encode.

    Parameters
    ----------
    obj : Any
        The object to encode

    Returns
    -------
    str
        The encoded object as a string
    """
    return jsonpickle.encode(obj,
                             keys=True,  # otherwise non-string keys will fail
                             unpicklable=True,  # just to be sure in case of future changes
                             make_refs=False,  # python ids are not preserved
                             warn=True,  # warn about unpicklable objects, safer development for now
                             max_iter=-1,  # no limit
                             indent=2,  # pretty print
                             separators=(",", ": "),  # pretty print
                             include_properties=False,  # can fail for class attributes
                             )


def decode(obj: str) -> Any:
    """
    A wrap with default parameters for jsonpickle.decode

    Parameters
    ----------
    obj : str
        The encoded object as a string

    Returns
    -------
    Any
        The decoded object
    """
    return jsonpickle.decode(obj,
                             keys=True,  # otherwise non-string keys will fail
                             on_missing='error')


# register all Handlers automatically when this module is imported
EnumHandler.handles(Enum)
# have to register all enum derived classes separate
if is_imported(Status):
    EnumHandler.handles(Status)
if is_imported(LogicCoupling):
    EnumHandler.handles(LogicCoupling)
if is_imported(Label):
    EnumHandler.handles(Label)
if is_imported(LebedevSphere):
    SimpleReprEvalHandler.handles(LebedevSphere)
if is_imported(PlaceHolderModelType):
    ProxyHandler.handles(PlaceHolderModelType)
ElementHandler.handles(ElementType)
AtomCollectionHandler.handles(AtomCollection)
jsonpickle_numpy.register_handlers()  # numpy objects



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""

from typing import List, Optional

from PySide2.QtWidgets import (
    QComboBox,
    QTreeView,
)
from PySide2.QtCore import QObject, Qt
from PySide2.QtGui import QStandardItem, QStandardItemModel, QCursor
from scine_heron.styling.delegates import CustomLightDelegate
import scine_heron.config as config


class BaseBox(QComboBox):
    """
    A combo box serving as a base class for all QComboBoxes in Heron and handling some basic formatting.
    """

    def __init__(self, parent: Optional[QObject] = None, header: Optional[str] = None) -> None:
        QComboBox.__init__(self, parent=parent)
        self.setCursor(QCursor(Qt.CursorShape.PointingHandCursor))

        tree_view = QTreeView()
        if config.MODE in config.LIGHT_MODES:
            tree_view.setItemDelegate(CustomLightDelegate())
        self.setView(tree_view)
        self.setModel(QStandardItemModel())
        self.view().setHeaderHidden(True)
        self.view().setItemsExpandable(False)
        self.view().setRootIsDecorated(False)

        if header is not None:
            parent_header = QStandardItem(header)
            parent_header.setEnabled(False)
            self.model().appendRow(parent_header)


class ComboBox(BaseBox):
    """
    A combo box that can be constructed with the given values that will be contained.
    """

    def __init__(self, parent: Optional[QObject], values: List[str], header: Optional[str] = None,
                 add_none: bool = False) -> None:
        """
        Construct the widget based on the class searcher

        Parameters
        ----------
        parent : QObject
            The parent widget
        values : List[str]
            The values within the combo box
        header : Optional[str], optional
            The non-selectable header in the combo box, by default no header
        add_none : bool, optional
            If 'None' should be added as an option to the combo box, by default False
        """
        super().__init__(parent=parent)

        if not values and not add_none:
            raise ValueError("No values given to ComboBox and add_none is False.")

        for name in values:
            child_name = name
            child_item = QStandardItem(name)
            self.model().appendRow(child_item)

        if add_none:
            child_name = 'none'
            child_item = QStandardItem(child_name)
            self.model().appendRow(child_item)

        self.setCurrentText(child_name)



__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""

# import typing
# from functools import partial
# from typing import Any, Callable, Dict, List, Tuple
#
# if typing.TYPE_CHECKING:
#     Signal = typing.Any
# else:
#     pass
#
#
# from PySide2.QtCore import QParallelAnimationGroup, QSize, Qt
# from PySide2.QtWidgets import (QComboBox, QDialog, QDockWidget, QFrame,
#                                QGridLayout, QHBoxLayout, QLabel, QLineEdit,
#                                QPushButton, QScrollArea, QSpinBox, QToolButton,
#                                QTreeWidget, QTreeWidgetItem, QVBoxLayout,
#                                QWidget, QDoubleSpinBox, QRadioButton, QCheckBox)
#
# from scine_heron.autocas.autocas_settings import AutocasSettings
# from scine_heron.autocas.basic_options import Interfaces
# from scine_heron.autocas.open_molcas_popup import OpenMolcasOptionsPopup
# from scine_heron.autocas.options_status_manager import OptionsStatusManager
# from scine_heron.autocas.signal_handler import SignalHandler
# from scine_heron.settings.settings import LabelsStyle, MoleculeStyle
# from scine_heron.status_manager import Status, StatusManager
#
# import os
#
#
# class OptionsWidget(QDockWidget):
#     def __init__(
#         self,
#         parent,
#         # options_status_manager: OptionsStatusManager,
#         signal_handler: SignalHandler,
#         autocas_settings: AutocasSettings,
#     ):
#         QDockWidget.__init__(self, "")
#
#         self.signal_handler = signal_handler
#         self.autocas_settings = autocas_settings
#
#         self.setFeatures(QDockWidget.NoDockWidgetFeatures)
#         self.__dockedWidget = QWidget(self)
#         self.setWidget(self.__dockedWidget)
#         self.__layout = QVBoxLayout()
#         self.__dockedWidget.setLayout(self.__layout)
#         self.__layout.setAlignment(Qt.AlignTop)
#
#         self.molecular_options = MolecularOptionsWidget(self, self.signal_handler, self.autocas_settings)
#         self.__layout.addWidget(self.molecular_options)
#         self.molecular_options.molecular_options()
#
#         self.molecular_options = MolecularOptionsWidget(self, self.signal_handler, self.autocas_settings)
#         self.__layout.addWidget(self.molecular_options)
#         self.molecular_options.molecular_options()
#
#         self.initial_orbital_options = InitialOrbitalOptionsWidget(self, self.signal_handler, self.autocas_settings)
#         self.__layout.addWidget(self.initial_orbital_options)
#         self.initial_orbital_options.initial_orbital_options()
#         # self.autocas_options()
#         # self.molcas_options()
#         # self.initial_orbital_options()
#         # self.initial_dmrg_options()
#         # self.final_calculation_options()
#
#     def basic_options(self):
#         pass
#
#
# class MolecularOptionsWidget(QWidget):
#     def __init__(self, parent, signal_handler, autocas_settings):
#         QWidget.__init__(self, parent)
#         self.layout = QGridLayout()
#         self.layout.setColumnMinimumWidth(0, 100)
#         self.layout.setColumnMinimumWidth(1, 100)
#         self.layout.setColumnMinimumWidth(2, 100)
#         self.layout.setColumnMinimumWidth(3, 100)
#         self.layout.setColumnMinimumWidth(4, 100)
#         self.layout.setColumnMinimumWidth(5, 200)
#         self.layout.setColumnMinimumWidth(6, 100)
#
#         self.layout.setRowMinimumHeight(0, 60)
#         self.layout.setRowMinimumHeight(1, 60)
#         self.layout.setRowMinimumHeight(2, 60)
#         self.layout.setRowMinimumHeight(3, 60)
#         self.layout.setRowMinimumHeight(4, 60)
#
#         self.basis_label = QLabel("Basis Set")
#         self.basis_label.setMaximumWidth(100)
#         self.basis_set_edit = QLineEdit(self)
#         self.basis_set_edit.setText("cc-pvdz")
#         self.basis_set_edit.setMaximumSize(100, 40)
#
#         self.charge_label = QLabel("Charge")
#         self.charge_label.setMaximumWidth(100)
#         self.charge_edit = QSpinBox()
#         self.charge_edit.setValue(0)
#         self.charge_edit.setMaximumSize(100, 40)
#
#         self.spin_multiplicity_label = QLabel("Spin Multiplicity")
#         self.spin_multiplicity_label.setMaximumWidth(200)
#         self.spin_multiplicity_edit = QSpinBox()
#         self.spin_multiplicity_edit.setValue(1)
#         self.spin_multiplicity_edit.setMinimum(1)
#         self.spin_multiplicity_edit.setMaximumWidth(100)
#
#         self.number_of_roots_label = QLabel("Number Of Roots")
#         self.number_of_roots_label.setMaximumWidth(100)
#         self.number_of_roots_edit = QSpinBox()
#         self.number_of_roots_edit.setValue(1)
#         self.number_of_roots_edit.setMinimum(1)
#
#         self.double_d_shell_label = QLabel("Include double d-shell")
#         self.double_d_shell_label.setMaximumWidth(100)
#         self.double_d_shell = QCheckBox()
#         self.double_d_shell.setChecked(True)
#         self.setLayout(self.layout)
#
#     def molecular_options(self):
#         # self.__layout.addWidget(QLabel("Molecular Settings"))
#         self.layout.addWidget(self.basis_label, 0, 0, 0, 1)
#         self.layout.addWidget(self.basis_set_edit, 0, 1, 0, 2)
#
#         self.layout.addWidget(self.charge_label, 0, 3, 0, 4)
#         self.layout.addWidget(self.charge_edit, 0, 4, 0, 5)
#
#         self.layout.addWidget(self.spin_multiplicity_label, 0, 5, 0, 6)
#         self.layout.addWidget(self.spin_multiplicity_edit, 0, 6, 0, 7)
#
#         # self.layout.addWidget(self.number_of_roots_label, 1, 0, 1, 1)
#         # self.layout.addWidget(self.number_of_roots_edit, 1, 1, 1, 2)
#         #
#         # self.layout.addWidget(self.double_d_shell_label, 1, 2, 1, 3)
#         # self.layout.addWidget(self.double_d_shell, 1, 3, 1, 4)
#         self.setLayout(self.layout)
#
#
# class InitialOrbitalOptionsWidget(QWidget):
#     def __init(self, parent, signal_handler, autocas_settings):
#         self.uhf = QCheckBox()
#         self.uhf.setChecked(False)
#
#     def initial_orbital_options(self):
#         self.__layout.addWidget(QLabel("Initial Orbital Settings"))
#
#         layout = QHBoxLayout()
#         sub_layout = QVBoxLayout()
#         sub_layout.addWidget(QLabel("Unrestricted"))
#         layout.addLayout(sub_layout)
#         sub_layout = QVBoxLayout()
#         sub_layout.addWidget(uhf)
#         layout.addLayout(sub_layout)
#         self.__layout.addLayout(layout)
#
#
# class MolcasOptionsWidget(QWidget):
#     def __init(self, parent, signal_handler, autocas_settings):
#         self.molcas_project_name = QLineEdit(self)
#         self.molcas_project_name.setText("autocas")
#
#         self.molcas_dump = QCheckBox()
#         self.molcas_dump.setChecked(True)
#
#         self.molcas_work_dir = QLineEdit(self)
#         self.molcas_work_dir.setText(os.getcwd() + "/test")
#
#         self.cholesky = QCheckBox()
#         self.cholesky.setChecked(True)
#
#     def molcas_options(self):
#         self.__layout.addWidget(QLabel("Molcas Settings"))
#
#         layout = QHBoxLayout()
#         sub_layout = QVBoxLayout()
#         sub_layout.addWidget(QLabel("Project Name"))
#         sub_layout.addWidget(QLabel("Enable Dump"))
#         sub_layout.addWidget(QLabel("Work Dir"))
#         sub_layout.addWidget(QLabel("Enable Cholesky decomposition"))
#         layout.addLayout(sub_layout)
#         sub_layout = QVBoxLayout()
#         sub_layout.addWidget(molcas_project_name)
#         sub_layout.addWidget(molcas_dump)
#         sub_layout.addWidget(molcas_work_dir)
#         sub_layout.addWidget(cholesky)
#         layout.addLayout(sub_layout)
#         self.__layout.addLayout(layout)
#
#
# class InitialDMRGOptions(QWidget):
#     def __init(self, parent, signal_handler, autocas_settings):
#         self.method = QComboBox()
#         self.method.addItems(["DMRG-CI", "DMRG-SCF"])
#
#         self.bond_dimension = QSpinBox()
#         self.bond_dimension.setMaximum(6000)
#         self.bond_dimension.setMinimum(1)
#         self.bond_dimension.setValue(250)
#
#         self.n_sweeps = QSpinBox()
#         self.n_sweeps.setValue(5)
#         self.n_sweeps.setMinimum(1)
#
#         self.fiedler_ordering = QCheckBox()
#         self.fiedler_ordering.setChecked(True)
#
#         self.large_cas = QCheckBox()
#         self.large_cas.setChecked(False)
#
#         self.n_large_cas_orbitals = QSpinBox()
#         self.n_large_cas_orbitals.setMinimum(1)
#         self.n_large_cas_orbitals.setValue(10)
#
#     def initial_dmrg_options(self):
#         self.__layout.addWidget(QLabel("Initial DMRG Settings"))
#
#         layout = QHBoxLayout()
#         sub_layout = QVBoxLayout()
#         sub_layout.addWidget(QLabel("Method"))
#         sub_layout.addWidget(QLabel("Bond dimension"))
#         sub_layout.addWidget(QLabel("Number of Sweeps"))
#         sub_layout.addWidget(QLabel("Fiedler Ordering"))
#         sub_layout.addWidget(QLabel("Enable Large CAS Protocol"))
#         sub_layout.addWidget(QLabel("Number of orbitals per sub-CAS"))
#         layout.addLayout(sub_layout)
#         sub_layout = QVBoxLayout()
#         sub_layout.addWidget(method)
#         sub_layout.addWidget(bond_dimension)
#         sub_layout.addWidget(n_sweeps)
#         sub_layout.addWidget(fiedler_ordering)
#         sub_layout.addWidget(large_cas)
#         sub_layout.addWidget(n_large_cas_orbitals)
#         layout.addLayout(sub_layout)
#         self.__layout.addLayout(layout)
#
#
# class FinalCalculationOptions(QWidget):
#     def __init(self, parent, signal_handler, autocas_settings):
#         self.method = QComboBox()
#         self.method.addItems(["DMRG-CI", "DMRG-SCF", "CASSCF", "CASCI"])
#
#         self.bond_dimension = QSpinBox()
#         self.bond_dimension.setMaximum(6000)
#         self.bond_dimension.setMinimum(1)
#         self.bond_dimension.setValue(2000)
#
#         self.n_sweeps = QSpinBox()
#         self.n_sweeps.setValue(20)
#         self.n_sweeps.setMinimum(1)
#
#         self.fiedler_ordering = QCheckBox()
#         self.fiedler_ordering.setChecked(True)
#
#         self.post_method = QComboBox()
#         self.post_method.addItems(["CASPT2", "NEVPT2"])
#
#         self.ipea = QDoubleSpinBox()
#         self.ipea.setValue(0.0)
#
#     def final_calculation_options(self):
#         self.__layout.addWidget(QLabel("Final Calculation Settings"))
#
#         layout = QHBoxLayout()
#         sub_layout = QVBoxLayout()
#         sub_layout.addWidget(QLabel("Method"))
#         sub_layout.addWidget(QLabel("Bond dimension"))
#         sub_layout.addWidget(QLabel("Number of Sweeps"))
#         sub_layout.addWidget(QLabel("Fiedler Ordering"))
#         sub_layout.addWidget(QLabel("Post CAS method"))
#         sub_layout.addWidget(QLabel("IPEA-Shift"))
#         layout.addLayout(sub_layout)
#         sub_layout = QVBoxLayout()
#         sub_layout.addWidget(method)
#         sub_layout.addWidget(bond_dimension)
#         sub_layout.addWidget(n_sweeps)
#         sub_layout.addWidget(fiedler_ordering)
#         sub_layout.addWidget(post_method)
#         sub_layout.addWidget(ipea)
#         layout.addLayout(sub_layout)
#         self.__layout.addLayout(layout)
#
#
# class AutoCASOptions(QWidget):
#     def __init(self, parent, signal_handler, autocas_settings):
#         self.plateau_values = QSpinBox(self)
#         self.plateau_values.setValue(10)
#
#         self.threshold_step = QDoubleSpinBox()
#         self.threshold_step.setValue(0.01)
#
#         self.weak_correlation_threshold = QDoubleSpinBox()
#         self.weak_correlation_threshold.setValue(0.02)
#
#         self.single_reference_threshold = QDoubleSpinBox()
#         self.single_reference_threshold.setValue(0.14)
#
#     def autocas_options(self):
#         self.__layout.addWidget(QLabel("AutoCAS Settings"))
#
#         plateau_values = QSpinBox(self)
#         plateau_values.setValue(10)
#
#         threshold_step = QDoubleSpinBox()
#         threshold_step.setValue(0.01)
#
#         weak_correlation_threshold = QDoubleSpinBox()
#         weak_correlation_threshold.setValue(0.02)
#
#         single_reference_threshold = QDoubleSpinBox()
#         single_reference_threshold.setValue(0.14)
#
#         layout = QHBoxLayout()
#         sub_layout = QVBoxLayout()
#         sub_layout.addWidget(QLabel("Required values for a plateau"))
#         sub_layout.addWidget(QLabel("Step size for a plateau"))
#         sub_layout.addWidget(QLabel("Weak correlation threshold"))
#         sub_layout.addWidget(QLabel("Single reference threshold"))
#         layout.addLayout(sub_layout)
#         sub_layout = QVBoxLayout()
#         sub_layout.addWidget(plateau_values)
#         sub_layout.addWidget(threshold_step)
#         sub_layout.addWidget(weak_correlation_threshold)
#         sub_layout.addWidget(single_reference_threshold)
#         layout.addLayout(sub_layout)
#         self.__layout.addLayout(layout)
#
#
# class PlotOptions(QWidget):
#     def __init(self, parent, signal_handler, autocas_settings):
#         self.basis_set_edit = QLineEdit(self)
#         self.basis_set_edit.setText("cc-pvdz")
#
#         self.charge_edit = QSpinBox()
#         self.charge_edit.setValue(0)
#         self.charge_edit.setMinimum(-100000)
#
#         self.spin_multiplicity_edit = QSpinBox()
#         self.spin_multiplicity_edit.setValue(1)
#         self.spin_multiplicity_edit.setMinimum(1)
#
#         self.number_of_roots_edit = QSpinBox()
#         self.number_of_roots_edit.setValue(1)
#         self.number_of_roots_edit.setMinimum(1)
#
#     def plot_options(self):
#         self.__layout.addWidget(QLabel("Initial Orbital Settings"))
#
#         layout = QHBoxLayout()
#         sub_layout = QVBoxLayout()
#         sub_layout.addWidget(QLabel("Basis Set"))
#         sub_layout.addWidget(QLabel("Charge"))
#         sub_layout.addWidget(QLabel("Spin Multiplicity"))
#         sub_layout.addWidget(QLabel("Number Of Roots"))
#         layout.addLayout(sub_layout)
#         sub_layout = QVBoxLayout()
#         sub_layout.addWidget(basis_set_edit)
#         sub_layout.addWidget(charge_edit)
#         sub_layout.addWidget(spin_multiplicity_edit)
#         sub_layout.addWidget(number_of_roots_edit)
#         layout.addLayout(sub_layout)
#         self.__layout.addLayout(layout)
#
#         # self.setFeatures(QDockWidget.NoDockWidgetFeatures)
#         #
#         # self.__dockedWidget = QWidget(self)
#         # self.setWidget(self.__dockedWidget)
#         #
#         # self.__layout = QVBoxLayout()
#         # self.__dockedWidget.setLayout(self.__layout)
#         # self.__layout.setAlignment(Qt.AlignTop)
#         #
#         # # self.__options_status_manager = options_status_manager
#         #
#         # self.__widgets_dict: Dict[str, Any] = {}
#         # self.__widget_height = 30
#         # self.__widget_width = 130
#         #
#         # self.__enabled = StatusManager(True)
#         #
#         # self.__add_interface_option(
#         #     "Interface",
#         #     [Interfaces.OpenMolcas.value],  # , Interfaces.Serenity.value],
#         #     self.__update_interface,
#         #     Interfaces.OpenMolcas.value,
#         #     enabled=self.__enabled,
#         # )
#         # self.__add_system_options_at_layout(enabled=self.__enabled)
#
#     # def button_clicked(self, s):
#     #     open_molcas_popup = OpenMolcasOptionsPopup(self)
#     #     open_molcas_popup.exec_()
#     #
#     # def __add_interface_option(
#     #     self,
#     #     option_name: str,
#     #     all_values: List[str],
#     #     update: Callable[[List[str], int], None],
#     #     default_value: str,
#     #     enabled: Status[bool],
#     # ) -> None:
#     #     combo_box = self.__add_combo_box_at_layout(option_name, option_name, all_values)
#     #     combo_box.currentIndexChanged.connect(partial(update, all_values))
#     #     combo_box.setCurrentIndex(all_values.index(default_value))
#     #     enabled.changed_signal.connect(combo_box.setEnabled)
#     #
#     #     self.interface_options = QPushButton("Advanced Options")
#     #     self.__layout.addWidget(self.interface_options)
#     #     self.interface_options.clicked.connect(self.button_clicked)
#     #
#     # def __add_system_options_at_layout(self, enabled: Status[bool]):
#     #     basis_set_edit = QLineEdit(self)
#     #     basis_set_edit.setText("cc-pvdz")
#     #
#     #     charge_edit = QSpinBox()
#     #     charge_edit.setValue(0)
#     #     charge_edit.setMinimum(-100000)
#     #
#     #     spin_multiplicity_edit = QSpinBox()
#     #     spin_multiplicity_edit.setValue(1)
#     #     spin_multiplicity_edit.setMinimum(1)
#     #
#     #     number_of_roots_edit = QSpinBox()
#     #     number_of_roots_edit.setValue(1)
#     #     number_of_roots_edit.setMinimum(1)
#     #
#     #     layout = QHBoxLayout()
#     #     sub_layout = QVBoxLayout()
#     #     sub_layout.addWidget(QLabel("Basis Set"))
#     #     sub_layout.addWidget(QLabel("Charge"))
#     #     sub_layout.addWidget(QLabel("Spin Multiplicity"))
#     #     sub_layout.addWidget(QLabel("Number Of Roots"))
#     #     layout.addLayout(sub_layout)
#     #     sub_layout = QVBoxLayout()
#     #     sub_layout.addWidget(basis_set_edit)
#     #     sub_layout.addWidget(charge_edit)
#     #     sub_layout.addWidget(spin_multiplicity_edit)
#     #     sub_layout.addWidget(number_of_roots_edit)
#     #     layout.addLayout(sub_layout)
#     #     self.__layout.addLayout(layout)
#     #
#     # def __update_interface(self, all_values: List[str], index: int) -> None:
#     #     """
#     #     Update molecule style.
#     #     """
#     #     self.__options_status_manager.interface = Interfaces(all_values[index])
#     #
#     # def __add_combo_box_at_layout(
#     #     self,
#     #     setting_name: str,
#     #     setting_key: str,
#     #     all_values: List[str],
#     # ) -> QComboBox:
#     #     """
#     #     Add QComboBox widget.
#     #     setting_name is a setting display name.
#     #     setting_key is a setting name in sparrow.
#     #     all_values is a list of valid values.
#     #     """
#     #     combo_box = QComboBox()
#     #     combo_box.addItems(all_values)
#     #     combo_box.setFixedSize(QSize(self.__widget_width, self.__widget_height + 1))
#     #
#     #     #   self.__widgets_dict[setting_key] = combo_box
#     #
#     #     layout = QHBoxLayout()
#     #     layout.addWidget(QLabel(setting_name))
#     #     layout.addWidget(combo_box)
#     #     self.__layout.addLayout(layout)
#     #
#     #     return combo_box
#     #
#     #
#     #
#     #
#     #
#     #
#     #
#     #
#     #
#     #
#     #
#     #
#     #
#     #
#     #
#     #
#     #
#     #
#     #
#     #
#     #
#     #
#     #
#     #
#     #
#     #
#     #
#     #
#     #
#     #
#     #
#
# # class SectionExpandButton(QPushButton):
# #     def __init__(self, item, text="", parent=None):
# #         super().__init__(text, parent)
# #         self.section = item
# #         self.clicked.connect(self.on_clicked)
# #
# #     def on_clicked(self):
# #         if self.section.isExpanded():
# #             self.section.setExpanded(False)
# #         else:
# #             self.section.setExpanded(True)
# #
# #
# # class OptionsWidget_bck(QDialog):
# #     def __init__(
# #         self,
# #         parent: QWidget,
# #         signal_handler: SignalHandler,
# #         autocas_settings: AutocasSettings,
# #     ):
# #         super().__init__()
# #         self.signal_handler = signal_handler
# #         self.autocas_settings = autocas_settings
# #         self.tree = QTreeWidget()
# #         self.tree.setHeaderHidden(True)
# #         layout = QVBoxLayout()
# #         layout.addWidget(self.tree)
# #         self.setLayout(layout)
# #         # self.tree.setIndentation(0)
# #
# #         self.sections: List[Tuple[QWidget, str]] = []
# #         self.define_sections()
# #         self.add_sections()
# #
# #     def add_sections(self):
# #         """adds a collapsible sections for every
# #         (title, widget) tuple in self.sections
# #         """
# #         for (title, widget) in self.sections:
# #             button1 = self.add_button(title)
# #             section1 = self.add_widget(button1, widget)
# #             button1.addChild(section1)
# #
# #     def define_sections(self):
# #         """reimplement this to define all your sections
# #         and add them as (title, widget) tuples to self.sections
# #         """
# #         widget = QFrame(self.tree)
# #         layout = QHBoxLayout(widget)
# #         layout.addWidget(
# #             OrbitalOption(self, self.signal_handler, self.autocas_settings)
# #         )
# #         # layout.addWidget(QLabel("Bla"))
# #         # layout.addWidget(QLabel("Blubb"))
# #         self.sections.append(("Hartree-Fock", widget))
# #         self.sections.append(("DMRG", widget))
# #
# #     def add_button(self, title):
# #         """creates a QTreeWidgetItem containing a button
# #         to expand or collapse its section
# #         """
# #         item = QTreeWidgetItem()
# #         self.tree.addTopLevelItem(item)
# #         self.tree.setItemWidget(item, 0, SectionExpandButton(item, text=title))
# #         return item
# #
# #     def add_widget(self, button, widget):
# #         """creates a QWidgetItem containing the widget,
# #         as child of the button-QWidgetItem
# #         """
# #         section = QTreeWidgetItem(button)
# #         section.setDisabled(True)
# #         self.tree.setItemWidget(section, 0, widget)
# #         return section
# #
# #
# # class OrbitalOption(QDockWidget):
# #     def __init__(self, parent, signal_handler, autocas_settings):
# #         super().__init__()
# #         self.setFeatures(QDockWidget.NoDockWidgetFeatures)
# #
# #         self.__dockedWidget = QWidget(self)
# #         self.setWidget(self.__dockedWidget)
# #
# #         self.__layout = QVBoxLayout()
# #         self.__dockedWidget.setLayout(self.__layout)
# #         self.__layout.setAlignment(Qt.AlignTop)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""

from PySide2.QtGui import QColor, QPalette
from PySide2.QtWidgets import (QHBoxLayout,
                               QTabWidget, QVBoxLayout, QWidget)

from scine_heron.autocas.entanglement_diagram import EntanglementDiagramWidget
from scine_heron.autocas.orbital_viewer_widget import OrbitalViewerWidget
from scine_heron.autocas.signal_handler import SignalHandler
from scine_heron.autocas.options import OptionsWidget
# from scine_heron.autocas.autocas_settings import AutocasSettings
# from scine_heron.autocas.mo_diagram import MODiagram
import scine_heron.utilities as utils


class LogoWidget(QWidget):
    def __init__(self):
        super().__init__()
        self.__layout = QHBoxLayout()
        palette = self.palette()
        palette.setColor(QPalette.Window, QColor("red"))
        palette.setColor(QPalette.Background, QColor("red"))
        self.setPalette(palette)
        self.setAutoFillBackground(True)

        self.setLayout(self.__layout)


class ResultWidget(QWidget):
    def __init__(
        self, parent: QWidget, signal_handler: SignalHandler, autocas_settings
    ):
        QWidget.__init__(self, parent)

        self.signal_handler = signal_handler
        self.autocas_settings = autocas_settings

        self.signal_handler.open_entanglement_widget_signal.connect(
            self.activate_entanglement
        )
        self.signal_handler.open_molecule_widget_signal.connect(self.activate_molecule)

        self.entanglement_diagram = EntanglementDiagramWidget(self, self.signal_handler)
        # handles all autocas options
        self.options = utils.vertical_scroll_area_wrap(OptionsWidget(self, self.signal_handler, self.autocas_settings))

        self.__layout = QVBoxLayout()
        self.__tabs = QTabWidget()
        self.__tabs.addTab(self.options, "Calculation Settings")
        self.__orbital_viewer_widget = OrbitalViewerWidget(self, self.signal_handler, self.autocas_settings)
        self.__tabs.addTab(self.__orbital_viewer_widget, "Orbital Viewer")
        self.__tabs.addTab(self.entanglement_diagram, "Orbital Entropies")
        # self.__tabs.addTab(self.entanglement_diagram, "Threshold Diagram")
        self.__tabs.setCurrentIndex(0)
        self.__layout.addWidget(self.__tabs)
        self.setLayout(self.__layout)

    def activate_molecule(self):
        self.__tabs.setCurrentIndex(0)

    def activate_entanglement(self):
        self.__tabs.setCurrentIndex(1)

    def activate_threshold(self):
        self.__tabs.setCurrentIndex(2)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""
from typing import Union

import scine_database as db


def get_compound_or_flask(a_id: db.ID, a_type: db.CompoundOrFlask, compound_collection: db.Collection,
                          flask_collection: db.Collection) -> Union[db.Compound, db.Flask]:
    if a_type == db.CompoundOrFlask.FLASK:
        return db.Flask(a_id, flask_collection)
    return db.Compound(a_id, compound_collection)


def aggregate_type_from_string(a_type_str: str) -> db.CompoundOrFlask:
    if a_type_str == "compound":
        return db.CompoundOrFlask.COMPOUND
    return db.CompoundOrFlask.FLASK



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""
"""
Provides the EditMoleculeWidget class.
"""

from silx.gui.widgets.PeriodicTable import PeriodicTable, PeriodicTableItem
from PySide2.QtGui import QIntValidator
from PySide2.QtCore import Qt
from PySide2.QtWidgets import (
    QWidget,
    QPushButton,
    QLabel,
    QGridLayout,
    QLineEdit,
)
from PySide2.QtGui import QKeySequence
from typing import Optional, Any, TYPE_CHECKING
if TYPE_CHECKING:
    Signal = Any
else:
    from PySide2.QtCore import Signal

from scine_heron.status_manager import StatusManager


class EditMoleculeWidget(QWidget):
    """
    Contains the tools to change the composition of the molecule.
    """

    addAtom = Signal(int)
    removeSelectedAtom = Signal()

    def __init__(self) -> None:
        super(EditMoleculeWidget, self).__init__()

        self.__enabled = StatusManager(True)

        self.__atomNumberLabel = QLabel("Atomic Number:")
        self.__atomNumberField = QLineEdit(self)
        self.__popUpPeriodicTableButton = QPushButton("Search...", self)
        self.__appendAtomButton = QPushButton("Add Atom", self)
        self.__atomNumberValidator = QIntValidator(1, 109, self)
        self.__atomNumberField.setValidator(self.__atomNumberValidator)
        self.__atomNumberField.setText("6")  # Default value
        self.__atomNumberField.textChanged.connect(  # pylint: disable=no-member
            self.__sanitize_atom_number
        )

        self.__removeAtomButton = QPushButton("Remove Selected Atom", self)

        self.__layout = QGridLayout()
        self.__layout.addWidget(self.__atomNumberLabel, 0, 0)
        self.__layout.addWidget(self.__atomNumberField, 0, 1)
        self.__layout.addWidget(self.__popUpPeriodicTableButton, 1, 0)
        self.__layout.addWidget(self.__appendAtomButton, 1, 1)
        self.__layout.addWidget(self.__removeAtomButton, 2, 0, 1, 2)
        self.__layout.setAlignment(Qt.AlignTop)

        self.setLayout(self.__layout)

        self.__appendAtomButton.clicked.connect(self.__appendAtom)  # pylint: disable=no-member
        self.__removeAtomButton.clicked.connect(self.__removeAtom)  # pylint: disable=no-member
        self.__removeAtomButton.setShortcut(QKeySequence("Del"))
        self.__popUpPeriodicTableButton.clicked.connect(  # pylint: disable=no-member
            self.__open_periodict_table_dialog
        )
        self.__periodicTableDialog: Optional[PeriodicTable] = None

    def set_enabled(self, enabled: bool) -> None:
        """
        Set widgets enabled.
        """
        self.__enabled.value = enabled

    def __set_atom_number(self, periodic_table_item: PeriodicTableItem) -> None:
        atomic_number = periodic_table_item.Z
        self.__atomNumberField.setText(str(atomic_number))
        if self.__periodicTableDialog is not None:
            self.__periodicTableDialog.close()
        self.addAtom.emit(int(atomic_number))

    def __open_periodict_table_dialog(self) -> None:
        self.__periodicTableDialog = PeriodicTable()
        self.__periodicTableDialog.setStyleSheet(
            """
            color: #000000;
            """
        )
        self.__periodicTableDialog.sigElementClicked.connect(self.__set_atom_number)
        self.__periodicTableDialog.show()

    def __sanitize_atom_number(self, _):
        if not self.__atomNumberField.hasAcceptableInput():
            text = self.__atomNumberField.text()
            if not text:
                return
            if text.isdigit():
                i = int(self.__atomNumberField.text())
                if i > 109:
                    self.__atomNumberField.setText("109")
                elif i < 1:
                    self.__atomNumberField.setText("1")
            else:
                self.__atomNumberField.setText("6")

    def __appendAtom(self) -> None:
        text = self.__atomNumberField.text()
        if text:
            atomic_number = self.__atomNumberField.displayText().strip()
            self.addAtom.emit(int(atomic_number))

    def __removeAtom(self) -> None:
        self.removeSelectedAtom.emit()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""
"""
Provides the Energy Profile widget.
"""

import numpy as np
import time

from scine_heron.utilities import (
    color_axis,
    color_figure,
    get_font,
    get_primary_line_color
)

from matplotlib.figure import Figure
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg
from datetime import datetime
from typing import List, Optional
from PySide2.QtCore import QSize
from PySide2.QtWidgets import (
    QDockWidget,
    QLabel,
    QWidget,
    QGridLayout,
    QSpinBox,
    QCheckBox,
)
from scine_heron.containers.combo_box import BaseBox
from scine_heron.energy_profile.energy_profile_status_manager import (
    EnergyProfileStatusManager,
)

from scine_heron.energy_profile.energy_profile_point import EnergyProfilePoint


class EnergyProfileWidget(QDockWidget):
    class Canvas(FigureCanvasQTAgg):  # type: ignore
        def __init__(
            self,
            parent: Optional[QWidget] = None,  # pylint: disable=unused-argument
            width: float = 10,
            height: float = 3,
        ):

            self.fig = Figure(figsize=(width, height))

            self.ax1 = self.fig.add_subplot(1, 1, 1, position=[0.15, 0.15, 0.75, 0.75])
            super(EnergyProfileWidget.Canvas, self).__init__(self.fig)

            self.energies: List[float] = []
            self.time_stamps: List[datetime] = []
            self.fixed_time: bool = True
            self.fixed_time_interval: int = 10
            self.last_update: float = float(time.time())
            self.energy_units: List[str] = ["kJ/mol", "Hartree"]
            self._energy_conversions: List[float] = [2625.5, 1.0]
            self.energy_unit: int = 0
            self.energy_type: int = 0

            (self.__line,) = self.ax1.plot([], [], color=get_primary_line_color())

            color_figure(self.fig)
            font = get_font()
            color_axis(self.ax1)
            self.ax1.set_title("Electronic Energy Profile", font)

            self.ax1.set_xlabel("Relative Time in s")
            if self.energy_type < 2:
                self.ax1.set_ylabel(
                    f"Relative energy in {self.energy_units[self.energy_unit]}"
                )
            else:
                self.ax1.set_ylabel(f"Energy in {self.energy_units[self.energy_unit]}")

            self.draw()

        def update_line(self) -> None:
            unit_conversion = self._energy_conversions[self.energy_unit]
            local_energies = self.energies
            local_time_stamps = self.time_stamps
            n = min(len(local_energies), len(local_time_stamps))
            if n < 2:
                return
            now = self.time_stamps[-1]
            if self.fixed_time:
                x = np.array(
                    [
                        (i - now).total_seconds()
                        for i in local_time_stamps[1:n]
                        if (now - i).total_seconds() < self.fixed_time_interval
                    ]
                )
                y = np.array(
                    [
                        i * unit_conversion
                        for i, j in zip(local_energies[1:n], local_time_stamps[1:n])
                        if (now - j).total_seconds() < self.fixed_time_interval
                    ]
                )
            else:
                x = np.array(
                    [
                        (i - now).total_seconds()
                        for i in local_time_stamps[1:n]
                    ]
                )
                y = np.array(local_energies[1:n]) * unit_conversion

            if self.energy_type == 0:
                ref = y[-1]
                y -= ref
            elif self.energy_type == 1:
                ref = np.min(y)
                y -= ref

            if len(y) > 1 and len(x) > 1:
                y_min = np.min(y)
                y_max = np.max(y)
                y_diff = y_max - y_min
                y_corr = max((50 / 2625.5) * unit_conversion - y_diff, 0.1 * y_diff)
                y_min -= 0.5 * y_corr
                y_max += 0.5 * y_corr
                x_min = np.min(x)
                x_max = np.max(x)
                x_diff = max(x_max - x_min, self.fixed_time_interval)
                x_min -= 0.1 * x_diff
                x_max += 0.1 * x_diff

                self.__line.set_data(x, y)
                self.__line.axes.axis([x_min, x_max, y_min, y_max])
            self.draw()

        def redraw_axis(self) -> None:
            if self.energy_type < 2:
                self.ax1.set_ylabel(
                    f"Relative energy in {self.energy_units[self.energy_unit]}"
                )
            else:
                self.ax1.set_ylabel(f"Energy in {self.energy_units[self.energy_unit]}")
            self.draw()

    def __init__(
        self,
        _: Optional[QWidget] = None,
        width: float = 10,
        height: float = 3,
    ):
        super(EnergyProfileWidget, self).__init__()
        self.setFeatures(QDockWidget.NoDockWidgetFeatures)

        self._inner_widget = QWidget(self)
        self.setWidget(self._inner_widget)

        self._layout = QGridLayout()
        self._layout.setRowMinimumHeight(0, 200)
        self._layout.setColumnMinimumWidth(0, 560)
        self._inner_widget.setLayout(self._layout)
        self.__widget_width = 170
        self.__widget_height = 30

        self.__canvas = self.Canvas(parent=self, width=width, height=height)
        EnergyProfileWidget.__canvas = self.__canvas

        self.energy_status_manager = EnergyProfileStatusManager()
        self.energy_status_manager.changed_signal.connect(self.update_energy_widget)
        self._layout.addWidget(self.__canvas, 0, 0, 6, 1)

        self._unit_label = QLabel()
        self._unit_label.setText("Unit of Energy:")
        self._unit_label.setFixedSize(QSize(self.__widget_width, self.__widget_height))
        self._layout.addWidget(self._unit_label, 1, 1, 1, 1)

        self._unit_cb = BaseBox()
        for e in self.__canvas.energy_units:
            self._unit_cb.addItem(e)
        self._unit_cb.currentIndexChanged.connect(self.update_energy_unit)  # pylint: disable=no-member
        self._unit_cb.setFixedSize(QSize(self.__widget_width, self.__widget_height))
        self._layout.addWidget(self._unit_cb, 2, 1, 1, 1)

        self._rel_unit_cb = BaseBox()
        # we assume later that every option after the first is a relative one
        for e in ["relative to now", "relative to lowest", "absolute"]:
            self._rel_unit_cb.addItem(e)
        self._rel_unit_cb.currentIndexChanged.connect(self.update_rel_energy_unit)  # pylint: disable=no-member
        self._rel_unit_cb.setFixedSize(QSize(self.__widget_width, self.__widget_height))
        self._layout.addWidget(self._rel_unit_cb, 3, 1, 1, 1)

        self._sliding_window_cbox = QCheckBox("Interval")
        self._sliding_window_cbox.setChecked(self.__canvas.fixed_time)
        self._sliding_window_cbox.toggled.connect(self.set_sliding_window)  # pylint: disable=no-member
        self._sliding_window_cbox.setFixedSize(
            QSize(self.__widget_width, self.__widget_height)
        )
        self._layout.addWidget(self._sliding_window_cbox, 4, 1, 1, 1)

        self._sliding_window_label = QLabel()
        self._sliding_window_label.setText("Interval in s:")
        self._sliding_window_label.setFixedSize(
            QSize(self.__widget_width, self.__widget_height)
        )
        self._layout.addWidget(self._sliding_window_label, 5, 1, 1, 1)

        self.__sliding_window_interval = QSpinBox()
        self.__sliding_window_interval.setMinimum(10)
        self.__sliding_window_interval.setMaximum(3600)
        self.__sliding_window_interval.setValue(self.__canvas.fixed_time_interval)
        self.__sliding_window_interval.valueChanged.connect(self.set_time_window)  # pylint: disable=no-member
        self.__sliding_window_interval.setFixedSize(
            QSize(self.__widget_width, self.__widget_height)
        )
        self._layout.addWidget(self.__sliding_window_interval, 6, 1, 1, 1)

    def set_sliding_window(self) -> None:
        self.__canvas.fixed_time = self._sliding_window_cbox.isChecked()

    def set_time_window(self, fixed_time_interval: int) -> None:
        self.__canvas.fixed_time_interval = fixed_time_interval

    def update_energy_widget(self, energy_profile: List[EnergyProfilePoint]) -> None:
        self.__canvas.energies.append(energy_profile[-1].energy)
        self.__canvas.time_stamps.append(energy_profile[-1].time_stamp)
        self.__canvas.last_update = float(time.time())
        self.__canvas.update_line()

    def update_energy_unit(self, new_unit: int) -> None:
        self.__canvas.energy_unit = new_unit
        self.__canvas.redraw_axis()
        self.__canvas.update_line()

    def update_rel_energy_unit(self, new_unit: int) -> None:
        self.__canvas.energy_type = new_unit
        self.__canvas.redraw_axis()
        self.__canvas.update_line()

    def reset(self) -> None:
        self.__canvas.energies = []
        self.__canvas.time_stamps = []
        self.__canvas.last_update = float(time.time())
        # Reset energy manager as well
        self.energy_status_manager.reset()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""
"""
Provides the ClassOptionsWidget class.
"""

from typing import Optional, List, TYPE_CHECKING, Tuple

from PySide2.QtCore import QObject
from PySide2.QtWidgets import (
    QLabel,
    QDialog,
    QWidget,
)

from scine_heron.containers.buttons import TextPushButton
from scine_heron.containers.combo_box_tab_widget import ComboBoxTabWidget
from scine_heron.containers.combo_box import ComboBox
from scine_heron.containers.layouts import VerticalLayout, HorizontalLayout
from scine_heron.dependencies.optional_import import is_imported, importer
from scine_heron.utilities import vertical_scroll_area_wrap

if TYPE_CHECKING:
    from scine_chemoton.steering_wheel.datastructures import GearOptions
    from scine_chemoton.gears.elementary_steps.trial_generator import TrialGenerator
    from scine_chemoton.gears import Gear
    from scine_heron.chemoton.class_searcher import ChemotonClassSearcher
    from scine_heron.chemoton.chemoton_widget import construct_all_possible_gears
else:
    GearOptions = importer("scine_chemoton.steering_wheel.datastructures", "GearOptions")
    TrialGenerator = importer("scine_chemoton.gears.elementary_steps.trial_generator", "TrialGenerator")
    Gear = importer("scine_chemoton.gears", "Gear")
    ChemotonClassSearcher = importer("scine_heron.chemoton.class_searcher", "ChemotonClassSearcher")
    construct_all_possible_gears = importer("scine_heron.chemoton.chemoton_widget", "construct_all_possible_gears")


class SingleGearOptionsWidget(QWidget):

    def __init__(self, gear_option: Tuple[Gear.Options, Optional[TrialGenerator.Options]],
                 name: str, parent: QWidget) -> None:
        from scine_heron.settings.dict_option_widget import DictOptionWidget

        super().__init__(parent=parent)
        self._parent = parent
        self._content_dict = {
            'Gear.Options': gear_option[0],
        }
        if gear_option[1] is not None:
            self._content_dict['TrialGenerator.Options'] = gear_option[1]

        content_widget = QWidget(parent=self)
        content_widget.setLayout(VerticalLayout([
            TextPushButton("Remove this option", lambda: self._parent.remove_tab(name), parent=self),
            DictOptionWidget(
                self._content_dict,
                parent=self,
                add_close_button=False,
                allow_additions=False,
                allow_removal=False,
                show_border=True,
            ),
        ]))
        self.setLayout(VerticalLayout([vertical_scroll_area_wrap(content_widget)]))


class GearOptionsWidget(ComboBoxTabWidget):
    """
    A popup widget for editing the options of a class or a dictionary.
    This class is basically a wrapper around DictOptionWidget.
    """

    def __init__(
            self,
            gear_options: Optional[GearOptions] = None,
            allowed_gears: Optional[List[Gear]] = None,
            parent: Optional[QObject] = None,
    ) -> None:
        super().__init__(parent=parent)
        if not is_imported(ChemotonClassSearcher):
            raise ImportError("ChemotonClassSearcher could not be imported.")

        self._layout.insertWidget(1, TextPushButton("Add new option to set", self._add_option, parent=self))
        if gear_options is None:
            gear_options = GearOptions()
        self._gear_options = gear_options
        if allowed_gears is None:
            allowed_gears = construct_all_possible_gears()
        self._allowed_gears = allowed_gears
        self._chemoton_gear_searcher = ChemotonClassSearcher(Gear)

    def get_options(self) -> GearOptions:
        return self._gear_options

    def _add_option(self) -> None:
        dialog = QDialog(self)
        dialog.setWindowTitle("Add option")
        gear_sele = ComboBox(parent=dialog, values=[g.name for g in self._allowed_gears], header="Gear", add_none=False)
        index_sele = ComboBox(parent=dialog, values=[str(i) for i in range(9)] + ['None'], header="Index")
        index_sele.setToolTip("Specify the index position of the gear starting at zero,"
                              "or 'None' to affect all gears of that type.")
        layout = VerticalLayout()
        layout.add_layouts([
            HorizontalLayout([QLabel("Gear: "), gear_sele]),
            HorizontalLayout([QLabel("Specify index position"), index_sele]),
        ])
        layout.add_widgets([
            TextPushButton("Cancel", dialog.reject),
            TextPushButton("Add", dialog.accept),
        ])
        dialog.setLayout(layout)
        if dialog.exec_():
            gear_name = gear_sele.currentText()
            # pick the first gear with that name
            gear = next(g for g in self._allowed_gears if g.name == gear_name)
            index_str = index_sele.currentText()
            if index_str.lower() == "none":
                index: Optional[int] = None
            else:
                index = int(index_str)
            self._gear_options += [(gear, index)]
            name = f"{gear.name}, {index}"
            single_option = SingleGearOptionsWidget(self._gear_options[(gear.name, index)], name, parent=self)
            self.add_tab(name, single_option)


class GearOptionsBuilderButtonWrapper(TextPushButton):

    def __init__(self, gear_options: Optional[GearOptions], allowed_gears: Optional[List[Gear]] = None,
                 parent: Optional[QObject] = None):
        super().__init__("Specify gear options", self.execute, parent=parent)
        if gear_options is None:
            gear_options = GearOptions()
        self._options = gear_options
        self._allowed_gears = allowed_gears

    def execute(self) -> None:
        builder = GearOptionsWidget(gear_options=self._options, allowed_gears=self._allowed_gears, parent=self)
        dialog = QDialog(self)
        dialog.setWindowTitle("Specify gear options")
        layout = VerticalLayout([
            builder,
            TextPushButton("Finished", dialog.accept),
        ])
        dialog.setLayout(layout)
        dialog.exec_()
        self._options = builder.get_options()

    def get_options(self) -> GearOptions:
        return self._options



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""

from dataclasses import dataclass, field
from inspect import isabstract
from typing import Any, List, Optional, Dict, Type, Tuple
from enum import Enum
from PySide2.QtWidgets import (
    QWidget,
    QGridLayout,
    QLabel,
    QVBoxLayout,
)

import numpy as np
# filters
from scine_chemoton.filters.aggregate_filters import (
    AggregateFilter,
    AggregateFilterAndArray,
    AggregateFilterOrArray,
)
from scine_chemoton.filters.reactive_site_filters import (
    ReactiveSiteFilter,
    ReactiveSiteFilterAndArray,
    ReactiveSiteFilterOrArray,
)
from scine_chemoton.filters.further_exploration_filters import (
    FurtherExplorationFilter,
    FurtherExplorationFilterAndArray,
    FurtherExplorationFilterOrArray,
)
from scine_chemoton.filters.elementary_step_filters import (
    ElementaryStepFilter,
    ElementaryStepFilterAndArray,
    ElementaryStepFilterOrArray
)
from scine_chemoton.filters.reaction_filters import (
    ReactionFilter,
    ReactionFilterAndArray,
    ReactionFilterOrArray
)
from scine_chemoton.filters.structure_filters import (
    StructureFilter,
    StructureFilterAndArray,
    StructureFilterOrArray,
)
# rule sets
from scine_chemoton.reaction_rules.distance_rules import (
    DistanceBaseRule,
    DistanceRuleAndArray,
    DistanceRuleOrArray,
)
from scine_chemoton.reaction_rules.polarization_rules import (
    PolarizationBaseRule,
    PolarizationRuleAndArray,
)

from scine_heron.chemoton.class_searcher import ChemotonClassSearcher
from scine_heron.chemoton.grouped_combo_box import GroupedComboBox
from scine_heron.containers.buttons import TextPushButton
from scine_heron.utilities import write_error_message, write_info_message


class _CombinationLogical(Enum):
    """
    Allows to control if the combination is a logical 'and' or 'or'
    """
    AND = "and"
    OR = "or"


@dataclass
class SelectionState:
    build_text: str = field(default="")
    class_instances_without_logicals: List[Any] = field(default_factory=list)
    current_open_logics: List[_CombinationLogical] = field(default_factory=list)
    last_closed_logical: Optional[_CombinationLogical] = field(default=None)


class ClassSelectionWidget(QWidget):
    """
    A class that allows to select classes of a certain type and combine them with
    'and' or 'or' logic, display the building process as a string based on the class
    names and then constructs the built object if requested.

    Notes
    -----
    * Currently assumes that the given class is not wanted to be selected, but only its subclasses.
    * The combination logic corresponding classes need to be added in the constructor
    to work for the corresponding classes (e.g., AggregateFilterAndArray for AggregateFilter),
    otherwise the logical combinations are simply forbidden.
    """

    def __init__(self, class_to_select: Type, class_description: str, parent: Optional[QWidget] = None):
        """
        Construct the widget based on a class type and a description.

        Parameters
        ----------
        class_to_select : Type
            The base class that we want to build
        class_description : str
            A well readable description of the class, mostly used for error messages.
        parent : Optional[QWidget], optional
            The parent widget, by default None
        """
        super().__init__(parent)
        self._class_description = class_description
        self._build_text: str = ""
        self._class_instances_without_logicals: List[Any] = []
        self._current_open_logics: List[_CombinationLogical] = []
        self._last_closed_logical: Optional[_CombinationLogical] = None

        self._logical_class_translation: Dict[Tuple[_CombinationLogical, Type], Type] = {
            (_CombinationLogical.AND, AggregateFilter): AggregateFilterAndArray,
            (_CombinationLogical.OR, AggregateFilter): AggregateFilterOrArray,
            (_CombinationLogical.AND, ReactiveSiteFilter): ReactiveSiteFilterAndArray,
            (_CombinationLogical.OR, ReactiveSiteFilter): ReactiveSiteFilterOrArray,
            (_CombinationLogical.AND, FurtherExplorationFilter): FurtherExplorationFilterAndArray,
            (_CombinationLogical.OR, FurtherExplorationFilter): FurtherExplorationFilterOrArray,
            (_CombinationLogical.AND, ElementaryStepFilter): ElementaryStepFilterAndArray,
            (_CombinationLogical.OR, ElementaryStepFilter): ElementaryStepFilterOrArray,
            (_CombinationLogical.AND, ReactionFilter): ReactionFilterAndArray,
            (_CombinationLogical.OR, ReactionFilter): ReactionFilterOrArray,
            (_CombinationLogical.AND, StructureFilter): StructureFilterAndArray,
            (_CombinationLogical.OR, StructureFilter): StructureFilterOrArray,
            (_CombinationLogical.AND, DistanceBaseRule): DistanceRuleAndArray,
            (_CombinationLogical.OR, DistanceBaseRule): DistanceRuleOrArray,
            (_CombinationLogical.AND, PolarizationBaseRule): PolarizationRuleAndArray,
        }
        skip_logic_buttons = not any(issubclass(class_to_select, k[1]) for k in self._logical_class_translation)

        black_list = []
        # add logical combinations to blacklist if they exist
        for logic in [_CombinationLogical.AND, _CombinationLogical.OR]:
            combination = self._logical_class_translation.get((logic, class_to_select))
            if combination is not None:
                black_list.append(combination)

        if class_to_select is ReactiveSiteFilter:
            black_list.append(FurtherExplorationFilter)  # FurtherFilter inherits from ReactiveSite

        self._searcher = ChemotonClassSearcher(class_to_select, black_list, avoid_exact_match=True)

        self._layout = QVBoxLayout(self)

        # buttons
        if not skip_logic_buttons:
            self._button_begin_and = TextPushButton("Begin logical And",
                                                    lambda: self._start_logical(_CombinationLogical.AND))
            self._button_begin_or = TextPushButton("Begin logical Or",
                                                   lambda: self._start_logical(_CombinationLogical.OR))
            self._button_end_and = TextPushButton("End logical And",
                                                  lambda: self._end_logical(_CombinationLogical.AND))
            self._button_end_or = TextPushButton("End logical Or",
                                                 lambda: self._end_logical(_CombinationLogical.OR))

            button_grid = QGridLayout()
            button_grid.addWidget(self._button_begin_and, 0, 0)
            button_grid.addWidget(self._button_end_and, 0, 1)
            button_grid.addWidget(self._button_begin_or, 1, 0)
            button_grid.addWidget(self._button_end_or, 1, 1)
            self._button_holder_widget = QWidget()
            self._button_holder_widget.setLayout(button_grid)
            self._layout.addWidget(self._button_holder_widget)

            # back button
            self._button_remove = TextPushButton("Remove last addition", self._remove_last)
            self._layout.addWidget(self._button_remove)

        # wipe button
        self._button_wipe = TextPushButton("Restart", self.restart)
        self._layout.addWidget(self._button_wipe)

        # class selection
        self._class_label = QLabel(f"{class_description}:", parent=self)
        self._class_box = GroupedComboBox(self, self._searcher)
        self._button_set_class = TextPushButton(f"Add {class_description}",
                                                lambda: self._add_instance(self._class_box.currentText().strip()),
                                                parent=self)
        self._layout.addWidget(self._class_label)
        self._layout.addWidget(self._class_box)
        self._layout.addWidget(self._button_set_class)

        # display
        self._display_label = QLabel("Current build:", parent=self)
        self._display = QLabel("", parent=self)
        self._display.setWordWrap(True)

        self._layout.addWidget(self._display_label)
        self._layout.addWidget(self._display)

        # verification button
        self._button_check = TextPushButton("Verify input", self.is_valid)  # type: ignore
        self._layout.addWidget(self._button_check)

        self.setLayout(self._layout)

    def _start_logical(self, logic: _CombinationLogical) -> None:
        """
        Start a logical bracket if sanity checks are passed.

        Parameters
        ----------
        logic : _CombinationLogical
            The logical operation that is started
        """
        if not self._current_open_logics and self._build_text:
            # we already have some instance(s), but no current logics to combine something, so it is too late to start
            write_error_message(f"You have already added a {self._class_description.lower()} "
                                f"and cannot start a logic operation only now")
            return
        if (logic, self._searcher.search_type) not in self._logical_class_translation:
            write_error_message(f"Cannot combine {self._searcher.search_type.__name__} with '{logic.value}' operation")
            return
        self._current_open_logics.append(logic)
        self._build_text_addition(f"{logic.value}(")

    def _end_logical(self, logic: _CombinationLogical) -> None:
        """
        End a logical bracket, if sanity checks are passed.

        Parameters
        ----------
        logic : _CombinationLogical
            The logical operation that is ended
        """
        if not self._current_open_logics or self._current_open_logics[-1] != logic:
            write_error_message(f"'{logic.value}' is not the currently open logical bracket.")
            return
        if self.build_text[-1] == "(":
            # we close, but just opened something, remove previous
            write_info_message("Removing opened logical operation")
            self._remove_last()
            self._current_open_logics.pop(-1)
        else:
            self._build_text_addition(")")
            self._last_closed_logical = self._current_open_logics.pop(-1)

    @property
    def build_text(self) -> str:
        """
        The description of the current build

        Returns
        -------
        str
            The build text
        """
        return self._build_text

    @build_text.setter
    def build_text(self, new_text: str):
        """
        Change the text of the current build, both internally and of the displayed box.

        Parameters
        ----------
        new_text : str
            The new build text.
        """
        self._build_text = new_text
        self._display.setText(self._build_text)

    def _build_text_addition(self, add: str):
        """
        Add something to the build text.

        Notes
        -----
        Does some additional padding based on brackets and commata.

        Parameters
        ----------
        add : str
            The added text.
        """
        if not self._build_text or self._build_text[-1] == "(" or add == ")":
            new_text = self._build_text + add
        else:
            new_text = self._build_text + f", {add}"
        self.build_text = new_text

    def _remove_last(self):
        """
        Removes the last logical or class that was added to the build.
        """
        chars = ['(', ')', ',']
        if not self._build_text:
            write_error_message("You have not added anything yet")
        elif self._build_text and self._build_text[-1] == ")":
            # just closed a logical, so simply remove bracket, and no class instances
            self.build_text = self._build_text[:-1]
            if self._last_closed_logical is None:
                write_error_message("Failed to remove last addition, wiping everything")
                self.restart()
            else:
                self._current_open_logics.append(self._last_closed_logical)
                # this can make problems with multiple deletes, but will likely just trigger a restart
                self._last_closed_logical = None
        # test for logical start at the end; -1 because of openeing bracket
        elif self.build_text[-4:-1] == _CombinationLogical.AND.value:
            self.build_text = self.build_text[:-4]
            self._current_open_logics.pop(-1)
        elif self.build_text[-3:-1] == _CombinationLogical.OR.value:
            self.build_text = self.build_text[:-3]
            self._current_open_logics.pop(-1)
        elif not self._class_instances_without_logicals:
            # we got no logical start and we also don't have a class instance, something is weird
            write_error_message("Failed to remove last addition, wiping everything")
            self.restart()
        elif len(self._class_instances_without_logicals) == 1 and not any(c in self._build_text for c in chars):
            # we got exactly one class and no logicals
            self._class_instances_without_logicals.pop(-1)
            self.build_text = ""
        else:
            # we should have a class instance as last thing to delete, delete point is either '(', ')', or ','
            # find the highest index with str.find()
            # str.find() gives first index, so let's reverse the string
            reversed_text = self._build_text[::-1]
            indices = [reversed_text.find(c) for c in chars]
            # find returns -1 if not in string, correct to infinity
            correct_indices = [i if i != -1 else np.inf for i in indices]
            char = chars[int(np.argmin(correct_indices))]
            new_text = char.join(self.build_text.split(char)[:-1])
            if char != ',':
                # keep the split character if it is not the comma
                new_text += char
            self._class_instances_without_logicals.pop(-1)
            self.build_text = new_text

    def restart(self) -> None:
        """
        Reset the building process to a clean built.
        """
        self.build_text = ""
        self._class_instances_without_logicals = []
        self._current_open_logics = []
        self._last_closed_logical = None

    def _add_instance(self, class_name: str):
        """
        Adds a new instance to the built, which might require user input for the
        construction arguments based on a pop-up.

        Parameters
        ----------
        class_name : str
            The name of the new class, that is looked up in the ChemotonClassSearcher.
        """
        from scine_heron.settings.class_options_widget import generate_instance_based_on_potential_widget_input

        if not self._current_open_logics and self._build_text:
            write_error_message(f"Cannot add an additional {self._class_description} without logical operations")
            return
        cls = self._searcher[class_name]
        instance = generate_instance_based_on_potential_widget_input(self, cls)
        if instance is None:
            return
        self._class_instances_without_logicals.append(instance)
        self._build_text_addition(class_name)

    def get_instance(self) -> Optional[Any]:
        """
        Construct the complete object based on the current built.

        Returns
        -------
        Optional[Any]
            The constructed class, None if not valid
        """
        if not self.build_text:
            write_info_message(f"No {self._class_description} has been specified")
            if isabstract(self._searcher.search_type):
                return None
            return self._searcher.search_type()
        if self._current_open_logics:
            write_error_message("We still have an open logic operation")
            return None
        if not any(bracket in self._build_text for bracket in ["(", ")"]):
            if len(self._class_instances_without_logicals) > 1:
                write_error_message("More than one instance was given, without any logical operations")
                return None
            return self._class_instances_without_logicals[0]

        first_logical = self._get_logical_in_3_chars(self._build_text[:3])

        grouping: Dict[_CombinationLogical, List[Any]] = {first_logical: []}  # data container
        current_lists = [grouping[first_logical]]  # structure to always point to currently filling structure

        # 'and' or ' or'
        last_3_chars = ' ' + first_logical.value if first_logical == _CombinationLogical.OR else first_logical.value
        start_index = 4 if first_logical == _CombinationLogical.AND else 3
        instance_index = 0
        for char in self._build_text[start_index:]:
            last_char = last_3_chars[-1]
            if char == ',' and last_char != ')':
                current_lists[-1].append(self._class_instances_without_logicals[instance_index])
                instance_index += 1
            elif char == '(':
                logical = self._get_logical_in_3_chars(last_3_chars)
                current_lists[-1].append({logical: []})
                current_lists.append(current_lists[-1][-1][logical])
            elif char == ')' and last_char != ')':
                current_lists[-1].append(self._class_instances_without_logicals[instance_index])
                instance_index += 1
                current_lists.pop(-1)
            last_3_chars = last_3_chars[1:] + char
        return self._convert_dictionary_to_instance(grouping)

    def _convert_dictionary_to_instance(self, grouping: Dict[_CombinationLogical, List[Any]]) -> Any:
        """
        Converts a dictionary of logicals and list of class instances to the final object.

        Parameters
        ----------
        grouping : Dict[_CombinationLogical, List[Any]]
            A dictionary describing the logical grouping.

        Returns
        -------
        Any
            The built object
        """
        if len(grouping.keys()) > 1:
            raise RuntimeError("internal error")
        key, values = list(grouping.items())[0]
        cls = self._logical_class_translation[(_CombinationLogical(key), self._searcher.search_type)]
        subcls = self._get_subclasses(values)
        return cls(subcls)

    def _get_subclasses(self, values: List[Any]) -> List[Any]:
        """
        A recursive method that allows to traverse the logical combinations to gather
        the complete list of subclasses of a logical combination class.

        Parameters
        ----------
        values : List[Any]
            The list of class instances or dictionaries to be interpreted.
            Dictionaries are evaluated in a recursive fashion.

        Returns
        -------
        List[Any]
            The list of all subclasses for a logical.
        """
        result = []
        for v in values:
            if isinstance(v, dict):
                if len(v.keys()) > 1:
                    raise RuntimeError("internal error")
                key, values = list(v.items())[0]
                cls = self._logical_class_translation[(_CombinationLogical(key), self._searcher.search_type)]
                subcls = self._get_subclasses(list(v.values())[0])
                result.append(cls(subcls))
            else:
                result.append(v)
        return result

    @staticmethod
    def _get_logical_in_3_chars(chars: str) -> _CombinationLogical:
        """
        Translate a string of 3 characters into a Logical Enum.

        Parameters
        ----------
        chars : str
            The string to be parsed

        Returns
        -------
        _CombinationLogical
            The logical described by the string

        Raises
        ------
        RuntimeError
            If string is not of length three or none of the implemented logicals
            are recognized.
        """
        if not len(chars) == 3:
            raise RuntimeError("Internal Error")
        if chars[:2] == _CombinationLogical.OR.value or chars[-2:] == _CombinationLogical.OR.value:
            return _CombinationLogical.OR
        if chars == _CombinationLogical.AND.value:
            return _CombinationLogical.AND
        raise RuntimeError(f"Internal Error: '{chars}'")

    def is_valid(self) -> bool:
        """
        If the current built is valid. This simply tested by trying to construct
        an object from the text with the associated sanity checks, no further test are done.

        Notes
        -----
        The result is also emitted as an info or error message.

        Returns
        -------
        bool
            If the build is valid
        """
        valid = self.get_instance() is not None
        if valid:
            write_info_message(f"Valid {self._class_description} has been specified")
        else:
            write_error_message(f"The {self._class_description} is not valid!")
        return valid

    def get_state(self) -> SelectionState:
        return SelectionState(self.build_text, self._class_instances_without_logicals,
                              self._current_open_logics, self._last_closed_logical)

    def set_state(self, state: SelectionState) -> None:
        self.build_text = state.build_text
        self._class_instances_without_logicals = state.class_instances_without_logicals
        self._current_open_logics = state.current_open_logics
        self._last_closed_logical = state.last_closed_logical



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""

import re
from typing import Any, List, Tuple, Optional

import numpy as np

from scine_heron.electronic_data.electronic_data import (Atom, ElectronicData,
                                                         GaussianOrbital,
                                                         MolecularOrbital)


class MoldenFileReader:
    """
    Provide molden file reader.
    """

    def read_molden(self, molden: str) -> ElectronicData:
        """
        This method read molden file and return ElectronicData.
        """
        atoms, gto, mo = self.__parse_molden(molden)
        parsed_atoms = self.__parse_atoms(atoms)
        self.__parse_gto(parsed_atoms, gto)
        mos = self.__parse_mo(mo)
        return ElectronicData(parsed_atoms, mos)

    def __parse_molden(self, molden: str) -> Tuple[List[str], List[str], List[str]]:
        """
        This method parse molden file.
        """

        is_atom_line = False
        is_gto_line = False
        is_mo_line = False

        atoms: List[str] = list()
        gto: List[str] = list()
        mo: List[str] = list()
        for line in molden.split("\n"):
            if re.match("^" + re.escape("[Atoms]"), line):
                is_atom_line = True
                is_gto_line = False
                is_mo_line = False
                continue
            elif re.match("^" + re.escape("[5D]"), line):
                is_atom_line = False
                is_gto_line = False
                is_mo_line = False
                continue
            elif re.match("^" + re.escape("[7F]"), line):
                is_atom_line = False
                is_gto_line = False
                is_mo_line = False
                continue
            elif re.match("^" + re.escape("[9G]"), line):
                is_atom_line = False
                is_gto_line = False
                is_mo_line = False
                continue
            elif re.match("^" + re.escape("[Charge] (Mullike)"), line):
                is_atom_line = False
                is_gto_line = False
                is_mo_line = False
                continue
            elif re.match("^" + re.escape("[GTO]"), line):
                is_atom_line = False
                is_gto_line = True
                is_mo_line = False
                continue
            elif re.match("^" + re.escape("[MO]"), line):
                is_atom_line = False
                is_gto_line = False
                is_mo_line = True
                continue

            if is_atom_line:
                atoms.append(line)

            elif is_gto_line:
                if line.startswith("["):
                    continue
                gto.append(line)

            elif is_mo_line:
                mo.append(line)

        return atoms, gto, mo

    def __parse_atoms(self, atoms: List[str]) -> List[Atom]:
        return [Atom.from_molden_line(line) for line in atoms]

    def __parse_mo(self, mo: List[str]) -> List[MolecularOrbital]:
        """
        This method parse MolecularOrbital.
        """
        mo_section_heads = self.__get_section_heads_and_positions(mo, r"\s?Sym=")
        mo_orbital_blocks = self.__split_by_sections(mo, mo_section_heads)
        return [
            MolecularOrbital.from_molden_file(block)
            for _, block in enumerate(mo_orbital_blocks)
        ]

    # def __stupid_gto_thing(self, gtos: List[str], atoms):
    #     re_filter = re.compile("[spdfgh]")
    #     j = 0
    #     stupid_list = []
    #     gto_list = []
    #     other_list = []
    #     atom_numbers = []

    #     for _, line in enumerate(gtos):

    #         # empty line
    #         if line.strip() == "":
    #             # ....
    #             gto_list.append(line)
    #             other_list.append(gto_list)
    #             gto_list = []
    #             j = 0
    #             if other_list == []:
    #                 pass
    #             else:
    #                 stupid_list.append(other_list)
    #             other_list = []
    #
    #         # found s p d f g h
    #         elif re_filter.match(line):
    #             if gto_list == []:
    #                 pass
    #             else:
    #                 other_list.append(gto_list)
    #             gto_list = [line]

    #         elif j == 0:
    #             atom_numbers.append(int(line.split()[0]))
    #             j = 1

    #         elif j == 1:
    #             gto_list.append(line)

    #     return stupid_list, atom_numbers

    def __new_atoms(self, gtos: List[str], atoms):
        re_filter = re.compile("[spdfgh]")
        gaussian_orbitals: List[GaussianOrbital] = []
        gaussian_coeffs: List[List[float]] = []
        j = 0
        atom = atoms[0]
        orbital_type: Optional[str] = None
        new_atoms = []
        for i, line in enumerate(gtos):
            # last line
            if i + 1 == len(gtos):
                assert orbital_type
                gaussian_orbitals.append(GaussianOrbital(orbital_type, gaussian_coeffs))
                atom.gaussian_orbitals = gaussian_orbitals
                new_atoms.append(atom)
            # found s p d f g h
            elif re_filter.match(line.lstrip()):
                if orbital_type:
                    gaussian_orbitals.append(
                        GaussianOrbital(orbital_type, gaussian_coeffs)
                    )
                orbital_type = line.split()[0]
                gaussian_coeffs = []
            # empty line
            elif line.strip() == "":
                assert orbital_type
                gaussian_orbitals.append(GaussianOrbital(orbital_type, gaussian_coeffs))
                atom.gaussian_orbitals = gaussian_orbitals
                new_atoms.append(atom)
                orbital_type = None
                j = 0
            # atom
            elif j == 0:
                gaussian_orbitals = []
                atom = atoms[int(line.split()[0]) - 1]
                j = 1
            # coeffs
            elif j == 1:
                gaussian_coeffs.append([float(line.split()[0]), float(line.split()[1])])
        return new_atoms

    def __parse_gto(self, atoms: List[Atom], gto: List[str]) -> None:
        """
        This method parse AtomicOrbitalsGTO.
        """
        atoms = self.__new_atoms(gto, atoms)
        i = 0
        for atom in atoms:
            i = i + 1
            atom.min_alpha = min(
                [
                    np.min(atom.gaussian_orbitals[i].alpha)
                    for i in range(len(atom.gaussian_orbitals))
                ]
            )

            for u in range(len(atom.gaussian_orbitals)):
                atom.sum_chi_step += atom.gaussian_orbitals[u].chi_step()

    # @staticmethod
    # def __parse_orbital_blocks(gto_block: List[Any]) -> List[GaussianOrbital]:
    #     gaussian_orbitals = []
    #     for orb_block in gto_block:
    #         try:
    #             orb_type, _, _ = orb_block[0].split()
    #         except ValueError:
    #             orb_type, _ = orb_block[0].split()

    #         gaussian_coeffs = [
    #             [float(x.replace("D", "E")) for x in coeff_line.split()]
    #             for coeff_line in orb_block[1:]
    #             if coeff_line not in ("", "\n")
    #         ]
    #         gaussian_orbitals.append(GaussianOrbital(orb_type, gaussian_coeffs))
    #     return gaussian_orbitals

    # @staticmethod
    # def __get_section_heads_and_positions_alternative(lines):
    #     sections = []
    #     j = 0
    #     for i, line in enumerate(lines):
    #         if line.strip() == "":
    #             sections.append([j, lines[j]])
    #             j += i + 1
    #     return sections

    @staticmethod
    def __get_section_heads_and_positions(
        lines: List[str], section_re: str
    ) -> List[Any]:
        re_filter = re.compile(section_re)
        sections = [[i, line] for i, line in enumerate(lines) if re_filter.match(line)]
        return sections

    @staticmethod
    def __split_by_sections(
        lines: List[str], section_heads_and_positions: List[Any]
    ) -> List[Any]:
        blocks = []
        for i in range(len(section_heads_and_positions)):
            line_nr = section_heads_and_positions[i][0]
            try:
                next_line_nr = section_heads_and_positions[i + 1][0] - 1
            except IndexError:
                next_line_nr = len(lines) - 1
            blocks.append(lines[line_nr: next_line_nr + 1])
        return blocks
